# ========================================================================
# Copyright 2018 Emory University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
import abc
import inspect
from types import SimpleNamespace
from typing import Optional, Sequence, List

import mxnet as mx
from mxnet import gluon, nd
from mxnet.ndarray import NDArray

__author__ = 'Jinho D. Choi'


class NLPModel(gluon.Block):
    """
    :class:`NLPModel` is an abstract class providing helper methods to implement a neural network model.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @abc.abstractmethod
    def forward(self, *args):
        raise NotImplementedError(
            '%s.%s()' %
            (self.__class__.__name__, inspect.stack()[0][3]))

    # ======================================== namespaces ====================

    @staticmethod
    def namespace_input_layer(
            row: int,
            col: int,
            dropout: float = 0.0) -> SimpleNamespace:
        """
        :param row: the row dimension of the input matrix.
        :param col: the column dimension of the input matrix.
        :param dropout: the dropout rate applied to the input matrix.
        :return: the namespace of (row, col, dropout) for the input layer.
        """
        return SimpleNamespace(col=col, row=row, dropout=dropout)

    @staticmethod
    def namespace_output_layer(dim: int) -> SimpleNamespace:
        """
        :param dim: the output dimension.
        :return: the namespace of (dim) for the output layer.
        """
        return SimpleNamespace(dim=dim)

    @staticmethod
    def namespace_fuse_conv_layer(
            filters: int,
            activation: str,
            dropout: float = 0.0) -> SimpleNamespace:
        """
        :param filters: the number of filters.
        :param activation: the activation function.
        :param dropout: the dropout applied to the output of this convolution.
        :return: the namespace of (filters, activation, dropout) for the fuse convolution layer.
        """
        return SimpleNamespace(
            filters=filters,
            activation=activation,
            dropout=dropout)

    @staticmethod
    def namespace_ngram_conv_layer(
            ngrams: Sequence[int],
            filters: int,
            activation: str,
            pool: str = None,
            dropout: float = 0.0) -> SimpleNamespace:
        """
        :param ngrams: the sequence of n-gram kernels applied to the convolutions.
        :param filters: the number of filters applied to each convolution.
        :param activation: the activation function applied to each convolution.
        :param dropout: the dropout rate applied to each convolution.
        :param pool: the pooling operation applied to each convolution (max|avg).
        :return: the namespace of (ngrams, filters, activation, pool, dropout) for the convolution layer.
        """
        return SimpleNamespace(
            ngrams=ngrams,
            filters=filters,
            activation=activation,
            pool=pool,
            dropout=dropout)

    @staticmethod
    def namespace_hidden_layer(
            dim: int,
            activation: str,
            dropout: float) -> SimpleNamespace:
        """
        :param dim: the dimension of the hidden layer.
        :param activation: the activation function applied to the hidden layer.
        :param dropout: the dropout rate applied to the output of the hidden layer.
        :return: the namespace of (dim, activation, dropout) for the hidden layer.
        """
        return SimpleNamespace(dim=dim, activation=activation, dropout=dropout)

    # ======================================== create_layers =================

    def init_input_layer(self, config: SimpleNamespace) -> SimpleNamespace:
        """
        :param config: the output of :meth:`FFNNModel.namespace_input_layer`.
        :return: the namespace of (dropout) for the input layer.
        """
        layer = SimpleNamespace(dropout=mx.gluon.nn.Dropout(config.dropout))

        with self.name_scope():
            self.__setattr__('input_dropout', layer.dropout)

        return layer

    def init_output_layer(self, config: SimpleNamespace) -> SimpleNamespace:
        """
        :param config: the output of :meth:`FFNNModel.namespace_output_layer`.
        :return: the namespace of (dense) for the output layer.
        """
        layer = SimpleNamespace(dense=mx.gluon.nn.Dense(config.dim))

        with self.name_scope():
            self.__setattr__('output', layer.dense)

        return layer

    def init_fuse_conv_layer(
            self,
            config: SimpleNamespace,
            input_col: int) -> SimpleNamespace:
        """
        :param config: the output of :meth:`FFNNModel.namespace_input_conv_layer`.
        :param input_col: the column dimension of the input matrix.
        :return: the namespace of (conv, dropout) for the fuse convolution layer.
        """
        layer = SimpleNamespace(
            conv=mx.gluon.nn.Conv2D(
                channels=config.filters, kernel_size=(
                    1, input_col), strides=(
                    1, input_col), activation=config.activation), dropout=mx.gluon.nn.Dropout(
                config.dropout))

        with self.name_scope():
            self.__setattr__('de_conv', layer.conv)
            self.__setattr__('de_dropout', layer.dropout)

        return layer

    def init_ngram_conv_layer(
            self,
            config: SimpleNamespace,
            input_row: int,
            input_col: int) -> List[SimpleNamespace]:
        """
        :param config: the output of :meth:`FFNNModel.namespace_ngram_conv_layer`.
        :param input_row: the row dimension of the input matrix.
        :param input_col: the column dimension of the input matrix.
        :return: the namespace of (conv, dropout, pool) for the n-gram convolution layer.
        """

        def pool(pool_type: str, n: int):
            if pool_type is None:
                return None
            p = mx.gluon.nn.MaxPool2D if pool_type == 'max' else mx.gluon.nn.AvgPool2D
            return p(pool_size=(n, 1), strides=(n, 1))

        def conv(n: int):
            layer = SimpleNamespace(
                conv=mx.gluon.nn.Conv2D(
                    channels=config.filters, kernel_size=(
                        n, input_col), strides=(
                        1, input_col), activation=config.activation), dropout=mx.gluon.nn.Dropout(
                    config.dropout), pool=pool(
                        config.pool, input_row - n + 1))

            with self.name_scope():
                self.__setattr__('ngram_conv_%d' % n, layer.conv)
                self.__setattr__('ngram_conv_%d_dropout' % n, layer.dropout)
                if layer.pool:
                    self.__setattr__('ngram_conv_%d_pool' % n, layer.pool)

            return layer

        return [conv(ngram) for ngram in config.ngrams]

    def init_hidden_layers(
            self,
            configs: Sequence[SimpleNamespace]) -> List[SimpleNamespace]:
        """
        :param configs: the sequence of outputs of :meth:`FFNNModel.namespace_hidden_layers`.
        :return: the namespace of (dense, dropout) for the hidden layers.
        """

        def hidden(n: int, c: SimpleNamespace) -> SimpleNamespace:
            layer = SimpleNamespace(
                dense=mx.gluon.nn.Dense(units=c.dim, activation=c.activation),
                dropout=mx.gluon.nn.Dropout(c.dropout))

            with self.name_scope():
                self.__setattr__('hidden_%d' % n, layer.dense)
                self.__setattr__('hidden_%d_dropout' % n, layer.dropout)

            return layer

        return [hidden(i, config) for i, config in enumerate(configs)]


class FFNNModel(NLPModel):
    """
    :class:`FFNNModel` implements a Feed-Forward Neural Network (FFNN) consisting of
    an input layer, n-gram convolution layers (optional), hidden layers (optional), and an output layer.
    This network accepts one input layer and returns one output layer.
    """

    def __init__(self,
                 input_config: SimpleNamespace,
                 output_config: SimpleNamespace,
                 fuse_conv_config: Optional[SimpleNamespace] = None,
                 ngram_conv_config: Optional[SimpleNamespace] = None,
                 hidden_configs: Optional[Sequence[SimpleNamespace]] = None,
                 **kwargs):
        """
        :param input_config: the configuration for the input layer; see :meth:`FFNNModel.namespace_input_layer`.
        :param output_config: the configuration for the output layer; see :meth:`FFNNModel.namespace_output_layer`.
        :param fuse_conv_config: the configuration for the fuse convolution layer; see :meth:`FFNNModel.namespace_fuse_conv_layer`.
        :param ngram_conv_config: the configuration for the n-gram convolution layer; see :meth:`FFNNModel.namespace_ngram_conv_layer`.
        :param hidden_configs: the configurations for the hidden layers; see :meth:`FFNNModel.namespace_hidden_layer`.
        """
        super().__init__(**kwargs)

        # initialize
        self._input = self.init_input_layer(input_config)

        if fuse_conv_config:
            self._fuse_conv = self.init_fuse_conv_layer(
                fuse_conv_config, input_config.col)
            col = fuse_conv_config.filters
        else:
            self._fuse_conv = None
            col = input_config.col

        self._ngram_convs = self.init_ngram_conv_layer(
            ngram_conv_config, input_config.row, col) if ngram_conv_config else None
        self._hiddens = self.init_hidden_layers(
            hidden_configs) if hidden_configs else None
        self._output = self.init_output_layer(output_config)

    def forward(self, x: NDArray) -> NDArray:
        # input layer
        x = self._input.dropout(x)

        # dimensionality reduction layer
        if self._fuse_conv:
            x = x.reshape((0, 1, x.shape[1], x.shape[2]))
            x = self._fuse_conv.dropout(self._fuse_conv.conv(x))

        # convolution layer
        if self._ngram_convs:
            x = mx.nd.transpose(
                x, (0, 3, 2, 1)) if self._fuse_conv else x.reshape(
                (0, 1, x.shape[1], x.shape[2]))
            t = [
                c.dropout(
                    c.pool(
                        c.conv(x))) if c.pool else c.dropout(
                    c.conv(x).reshape(
                        (0, -1))) for c in self._ngram_convs]
            x = nd.concat(*t, dim=1)

        # hidden layers
        if self._hiddens:
            for h in self._hiddens:
                x = h.dense(x)
                x = h.dropout(x)

        # output layer
        output = self._output.dense(x)
        return output



# ========================================================================
# Copyright 2017 Emory University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
import abc
import inspect
import logging
from typing import List, Optional, Union, Sequence

import fastText
import numpy as np
from gensim.models import KeyedVectors
from mxnet import nd
from mxnet.ndarray import NDArray

from elit.util.structure import TOK, Document

__author__ = 'Jinho D. Choi'


class LabelMap:
    """
    :class:`LabelMap` provides the mapping between string labels and their unique integer IDs.
    """

    def __init__(self):
        self.index_map = {}
        self.labels = []

    def __len__(self):
        return len(self.labels)

    def __str__(self):
        return str(self.index_map)

    def add(self, label: str) -> int:
        """
        :param label: the label.
        :return: the class ID of the label.

        Adds the label to this map and assigns its class ID if not already exists.
        """
        idx = self.cid(label)
        if idx < 0:
            idx = len(self.labels)
            self.index_map[label] = idx
            self.labels.append(label)
        return idx

    def get(self, cid: int) -> str:
        """
        :param cid: the class ID.
        :return: the label corresponding to the class ID.
        """
        return self.labels[cid]

    def cid(self, label: str) -> int:
        """
        :param label: the label.
        :return: the class ID of the label if exists; otherwise, -1.
        """
        return self.index_map.get(label, -1)

    def argmax(self, scores: Union[np.ndarray, NDArray]) -> str:
        """
        :param scores: the prediction scores of all labels.
        :return: the label with the maximum score.
        """
        if self.__len__() < len(scores): scores = scores[:self.__len__()]
        n = nd.argmax(scores, axis=0).asscalar() if isinstance(scores, NDArray) else np.argmax(scores)
        return self.get(int(n))


# ======================================== Vector Space Models ========================================

class VectorSpaceModel(abc.ABC):
    """
    :class:`VectorSpaceModel` is an abstract class to implement vector space models.

    Abstract methods to be implemented:
      - :meth:`VectorSpaceModel.emb`
    """

    def __init__(self, dim: int):
        """
        :param dim: the dimension of each embedding.
        """
        self.dim = dim

        # for zero padding
        self.pad = np.zeros(dim).astype('float32')

    @abc.abstractmethod
    def emb(self, value: str) -> np.ndarray:
        """
        :param value: the value (e.g., a word) to retrieve the embedding for.
        :return: the embedding of the value.

        This is an auxiliary function for :meth:`VectorSpaceModel.embedding`.
        """
        raise NotImplementedError('%s.%s()' % (self.__class__.__name__, inspect.stack()[0][3]))

    def embedding(self, value: Optional[str] = None) -> np.ndarray:
        """
        :param value: the value (e.g., a word) to retrieve the embedding for.
        :return: the embedding of the value; if the value is ``None``, the zero embedding.
        """
        return self.emb(value) if value is not None else self.pad

    def embedding_list(self, values: Sequence[str]) -> List[np.ndarray]:
        """
        :param values: the sequence of values (e.g., words).
        :return: the list of embeddings for the corresponding values.
        """
        return [self.embedding(value) for value in values]

    def sentence_embedding_list(self, document: Document, key: str = TOK) -> List[List[np.ndarray]]:
        """
        :param document: the input document.
        :param key: the key to the values in each sentence to retrieve the embeddings for.
        :return: the list of embedding lists, where each embedding list contains the embeddings for the values in the corresponding sentence.
        """
        return [self.embedding_list(s[key]) for s in document]

    def embedding_matrix(self, values: Sequence[str], maxlen: int) -> List[np.ndarray]:
        """
        :param values: the sequence of values (e.g., words).
        :param maxlen: the maximum length of the output list;
                       if ``> len(values)``, the bottom part of the matrix is padded with zero embeddings;
                       if ``< len(values)``, embeddings of the exceeding values are discarded from the resulting matrix.
        :return: the matrix where each row is the embedding of the corresponding value.
        """
        return [self.embedding(values[i]) if i < len(values) else self.pad for i in range(maxlen)]


class FastText(VectorSpaceModel):
    """
    :class:`FastText` is a vector space model whose embeddings are trained by `FastText <https://github.com/facebookresearch/fastText>`_.
    """

    def __init__(self, filepath: str):
        """
        :param filepath: the path to the binary file containing a word embedding model trained by FastText.
        :type filepath: str
        """
        logging.info('FastText')
        logging.info('- model: %s' % filepath)
        self.model = fastText.load_model(filepath)
        dim = self.model.get_dimension()
        super().__init__(dim)
        logging.info('- vocab = %d, dim = %d' % (len(self.model.get_words()), dim))

    def emb(self, value: str) -> np.ndarray:
        return self.model.get_word_vector(value)


class Word2Vec(VectorSpaceModel):
    """
    :class:`Word2Vec` is a vector space model whose embeddings are trained by `Word2Vec <https://github.com/tmikolov/word2vec>`_.
    """

    def __init__(self, filepath: str):
        """
        :param filepath: the path to the binary file containing a word embedding model trained by Word2Ved.
        :type filepath: str
        """
        logging.info('Word2Vec')
        logging.info('- model: %s' % filepath)
        self.model = KeyedVectors.load(filepath) if filepath.lower().endswith('.gnsm') else KeyedVectors.load_word2vec_format(filepath, binary=True)
        dim = self.model.syn0.shape[1]
        super().__init__(dim)
        logging.info('- vocab = %d, dim = %d' % (len(self.model.vocab), dim))

    def emb(self, value: str) -> np.ndarray:
        vocab = self.model.vocab.get(value, None)
        return self.model.syn0[vocab.index] if vocab else self.pad


# class GloVe(VectorSpaceModel):
#     def __init__(self, filepath: str):
#         """
#         :param filepath: the path to the binary file containing word embeddings.
#         :type filepath: str
#         """
#         super().__init__(0)
#         raise NotImplementedError('%s.%s()' % (self.__class__.__name__, inspect.stack()[0][3]))
#
#     def emb(self, value: str) -> np.ndarray:
#         return self.pad


class Position2Vec(VectorSpaceModel):
    """
    :class:`Position2Vec` is a vector space model that generates position embeddings
    indicating the positions of tokens in a sentence (e.g., first token, last token).

    Supported method:
      - :meth:`Position2Vec.sentence_embedding_list`
    """
    _FST = np.array([1, 0]).astype('float32')  # the first word
    _LST = np.array([0, 1]).astype('float32')  # the last word
    _ANY = np.array([0, 0]).astype('float32')  # any other word

    def __init__(self):
        super().__init__(2)

    # override
    def emb(self, value: str) -> np.ndarray:
        """
        Unsupported (any token-based embeddings are not supported in this vector space model).
        """
        raise AttributeError('%s.%s()' % (self.__class__.__name__, inspect.stack()[0][3]))

    # override
    def sentence_embedding_list(self, document: Document, key: str = None) -> List[List[np.ndarray]]:
        """
        :param document: the input document.
        :param key: ``None``.
        :return: the list of embedding lists, where each embedding list contains the position embeddings of the corresponding sentence.
        """

        def aux(sentence):
            size = len(sentence)
            return [self._FST if i == 0 else self._LST if i + 1 == size else self._ANY for i in range(size)]

        return [aux(s) for s in document]


# ======================================== Embeddings ========================================


def x_extract(index: int, offset: int, embedding_list: Sequence[np.ndarray], pad: np.ndarray) -> np.ndarray:
    """
    :param index: the center index of the embedding to be retrived.
    :param offset: the offset to the center index.
    :param embedding_list: the sequence of embeddings.
    :param pad: the zero-pad embedding.
    :return: the (index + offset)'th embedding if available; otherwise, the zero-pad embedding.
    """
    i = index + offset
    return embedding_list[i] if 0 <= i < len(embedding_list) else pad



# ========================================================================
# Copyright 2018 Emory University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
import bisect
import glob
import json
import logging
from typing import List, Dict, Callable, Sequence

from elit.state import NLPState
from elit.util.structure import Sentence, TOK, Document, to_gold

__author__ = "Jinho D. Choi, Gary Lai"


# ======================================== Readers ========================================

def tsv_reader(filepath: str, cols: Dict[str, int], key: str = None) -> List[Document]:
    documents = []
    wc = sc = 0

    if TOK not in cols:
        raise ValueError('The column index of "%s" must be specified' % TOK)

    if key is not None:
        if key in cols:
            cols = cols.copy()
            cols[to_gold(key)] = cols.pop(key)
        else:
            raise ValueError('Key mismatch: %s is not a key in %s' % (key, str(cols)))

    logging.info('Reading')
    logging.info('- filepath: %s' % filepath)

    for filename in glob.glob(filepath):
        fin = open(filename)
        sentences = []
        fields = {k: [] for k in cols.keys()}

        for line in fin:
            if line.startswith('#'): continue
            l = line.split()

            if l:
                for k, v in fields.items():
                    v.append(l[cols[k]])
            elif len(fields[TOK]) > 0:
                wc += len(fields[TOK])
                sentences.append(Sentence(fields))
                fields = {k: [] for k in cols.keys()}

        if len(fields[TOK]) > 0:
            wc += len(fields[TOK])
            sentences.append(Sentence(fields))

        fin.close()
        sc += len(sentences)
        documents.append(Document(sen=sentences))

    logging.info('- dc = %d, sc = %d, wc = %d' % (len(documents), sc, wc))
    return documents


def json_reader(filepath: str) -> List[Document]:
    # TODO: update this to accept any format (see tsv_reader)
    documents = []
    dc = wc = sc = 0

    logging.info('Reading')
    logging.info('- filepath: %s' % filepath)

    for filename in glob.glob('{}/*.json'.format(filepath)):
        assert filename.endswith('.json')
        with open(filename) as f:
            docs = json.load(f)
            for doc in docs:
                sentences = []
                for sen in doc['sen']:
                    wc += len(sen['tok'])
                    sentences.append(Sentence(sen))
                sc += len(sentences)
                documents.append(Document(sen=sentences))
            dc += len(documents)
    logging.info('- dc = %d, sc = %d, wc = %d' % (dc, sc, wc))
    return documents


def pkl(filepath):
    return filepath + '.pkl'


def gln(filepath):
    return filepath + '.gln'


def group_states(docs: Sequence[Document], create_state: Callable[[Document], NLPState], maxlen: int = -1) -> List[NLPState]:
    """
    Groups sentences into documents such that each document consists of multiple sentences and the total number of words
    across all sentences within a document is close to the specified maximum length.
    :param docs: a list of documents.
    :param create_state: a function that takes a document and returns a state.
    :param maxlen: the maximum number of words; if max_len < 0, it is inferred by the length of the longest sentence.
    :return: list of states, where each state roughly consists of the max_len number of words.
    """

    def aux(i):
        ls = d[keys[i]]
        t = ls.pop()
        document.sentences.append(t)
        if not ls: del keys[i]
        return len(t)

    # key = length, value = list of sentences with the key length
    d = {}

    for doc in docs:
        for sen in doc.sentences:
            d.setdefault(len(sen), []).append(sen)

    keys = sorted(list(d.keys()))
    if maxlen < 0:
        maxlen = keys[-1]

    states = []
    document = Document()
    wc = maxlen - aux(-1)

    while keys:
        idx = bisect.bisect_left(keys, wc)
        if idx >= len(keys) or keys[idx] > wc:
            idx -= 1
        if idx < 0:
            states.append(create_state(document))
            document = Document()
            wc = maxlen - aux(-1)
        else:
            wc -= aux(idx)

    if document:
        states.append(create_state(document))

    return states



# ========================================================================
# Copyright 2018 Emory University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
import abc
import collections
import inspect
import random
from itertools import islice
from typing import Sequence, Union, List, Tuple

import numpy as np

from elit.state import NLPState

__author__ = 'Jinho D. Choi'


class NLPIterator(collections.Iterator):
    """
    :class:`NLPIterator` generates batches of instances from the input states that are iterable.
    """

    def __init__(self, states: Sequence[NLPState], label: bool):
        """
        :param states: the sequence of input states.
        :param label: if ``True``, each instance is ``(x, y)``; otherwise, it is just ``x``.
        """
        self.states = states
        self.label = label
        self.x1 = True

    @abc.abstractmethod
    def __iter__(self) -> 'NLPIterator':
        raise NotImplementedError('%s.%s()' % (self.__class__.__name__, inspect.stack()[0][3]))

    @abc.abstractmethod
    def __next__(self):
        raise NotImplementedError('%s.%s()' % (self.__class__.__name__, inspect.stack()[0][3]))

    def __str__(self):
        return self.__class__.__name__

    @abc.abstractmethod
    def process(self, *args):
        """
        :param args: arguments to be applied to the input states; it must include at least one argument.

        Processes through the states and assigns the parameters accordingly.
        """
        raise NotImplementedError('%s.%s()' % (self.__class__.__name__, inspect.stack()[0][3]))


class BatchIterator(NLPIterator):
    """
    :class:`BatchIterator` generates all batches of instances in the constructor, and iterates through the generated batches.
    This is useful when all instances are independent to one another.
    """

    def __init__(self, states: Sequence[NLPState], batch_size: int, shuffle: bool, label: bool, **kwargs):
        """
        :param states: the input states.
        :param batch_size: the max number of instances per batch.
        :param shuffle: if ``True``, shuffle instances for every epoch, in which case, :meth:`BatchIterator.process` cannot be called.
        :param label: if ``True``, each instance is ``(x, y)``; otherwise, it is just ``x``.
        :param kwargs: custom parameters to initialize each state.
        """
        super().__init__(states, label)
        batches = []
        batch = []
        count = 0

        for state in states:
            for x, y, in state:
                if label: batch.append((x, y))
                else: batch.append(x)
                count += 1

                if count == batch_size:
                    batches.append(batch)
                    batch = []
                    count = 0

        if batch: batches.append(batch)
        self._batches = batches
        self._shuffle = shuffle
        self._kwargs = kwargs
        self._state_begin = 0
        self._iter = 0

        x = batches[0][0]
        if isinstance(x, tuple): x = x[0]
        self.x1 = not isinstance(x, tuple)

    # override
    def __iter__(self) -> 'BatchIterator':
        if self._shuffle:
            random.shuffle(self._batches)
        else:
            for state in self.states: state.init(**self._kwargs)

        self._state_begin = 0
        self._iter = 0
        return self

    # override
    def __next__(self) -> List[Union[np.ndarray, Tuple[np.ndarray, int]]]:
        if self._iter >= len(self._batches): raise StopIteration
        batch = self._batches[self._iter]
        self._iter += 1
        return batch

    # override
    def process(self, *args):
        if self._shuffle: return -1
        o = len(args) == 1
        i = 0

        for state in islice(self.states, self._state_begin, None):
            while state.has_next():
                if i == len(args[0]): return
                if o: state.process(args[0][i])
                else: state.process(*[arg[i] for arg in args])
                i += 1

            self._state_begin += 1


class SequenceIterator(NLPIterator):
    """
    :class:`SequenceIterator` dynamically generates batches of instances during iteration, where each instance is extracted from a different input state.
    This is useful when the output of each state needs to be passed onto its following states.

    Important: :meth:`SequenceIterator.process` must be called at each iteration.
    """

    def __init__(self, states: Sequence[NLPState], batch_size: int, shuffle: bool, label: bool, **kwargs):
        """
        :param states: the input states.
        :param batch_size: the max number of instances per batch.
        :param shuffle: if ``True``, shuffle instances for every epoch.
        :param label: if ``True``, each instance is ``(x, y)``; otherwise, it is just ``x``.
        :param kwargs: custom parameters to initialize each state.
        """
        super().__init__(states, label)
        self._batch_size = batch_size
        self._shuffle = shuffle
        self._kwargs = kwargs
        self._states = None
        self._state_begin = 0
        self._iter = 0

    def __iter__(self) -> 'SequenceIterator':
        for state in self.states: state.init(**self._kwargs)

        self._states = list(self.states)
        if self._shuffle: random.shuffle(self._states)
        self.x1 = not isinstance(self.states[0].x, tuple)
        self._state_begin = 0
        self._iter = 0
        return self

    def __next__(self) -> List[Union[np.ndarray, Tuple[np.ndarray, int]]]:
        if not self._states: raise StopIteration
        batch = [(state.x, state.y) if self.label else state.x for state in islice(self._states, self._iter, self._iter + self._batch_size)]
        self._iter += len(batch)
        return batch

    def process(self, *args):
        o = len(args) == 1
        end = self._state_begin + len(args[0])

        for i, state in enumerate(islice(self._states, self._state_begin, end)):
            if o: state.process(args[0][i])
            else: state.process(*[arg[i] for arg in args])

        self._state_begin = end

        if end >= len(self._states):
            self._states = [state for state in self._states if state.has_next()]
            if self._shuffle: random.shuffle(self._states)
            self._state_begin = 0
            self._iter = 0



# ========================================================================
# Copyright 2017 Emory University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
import string

__author__ = 'Jinho D. Choi'


def is_range(c, begin, end):
    return begin <= c <= end


def is_single_quote(c):
    return c in {'\'', '`'} or is_range(c, u'\u2018', u'\u201B')


def is_double_quote(c):
    return c == '"' or is_range(c, u'\u201C', u'\u201F')


def is_left_bracket(c):
    return c in {'(', '{', '[', '<'}


def is_right_bracket(c):
    return c in {')', '}', ']', '>'}


def is_bracket(c):
    return is_left_bracket(c) or is_right_bracket(c)


def is_hyphen(c):
    return c == '-' or is_range(c, u'\u2010', u'\u2014')


def is_arrow(c):
    return is_range(c, u'\u2190', u'\u21FF') or is_range(c, u'\u27F0', u'\u27FF') or is_range(c, u'\u2900', u'\u297F')


def is_currency(c):
    return c == '$' or is_range(c, u'\u00A2', u'\u00A5') or is_range(c, u'\u20A0', u'\u20CF')


def is_final_mark(c):
    return c in {'.', '?', '!', u'\u203C'} or is_range(c, u'\u2047', u'\u2049')


def is_punct(c):
    return c in string.punctuation


def collapse_digits(s):
    def get(i, c):
        if i+1 < len(s) and digits[i+1]:
            if is_currency(c) or c in {'.', '-', '+', '#'}:
                return ''
            if i-1 >= 0 and digits[i-1] and (is_hyphen(c) or c in {',', ':', '/', '='}):
                return ''
        elif i-1 >= 0 and digits[i-1]:
            if c == '%':
                return ''
        elif digits[i]:
            if t and t[0] == '0':
                return ''

        t[0] = '0' if digits[i] else c
        return t[0]

    t = ['']
    digits = [c.isdigit() for c in s]
    return ''.join([get(i, c) for i, c in enumerate(s)])



# ========================================================================
# Copyright 2018 Emory University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
import abc
from typing import Any

from elit.component import Component
from elit.util.string import *
from elit.util.structure import SEN_ID, TOK, OFF

__author__ = "Gary Lai"


class Segmenter(Component):

    def init(self, **kwargs):
        pass

    def load(self, model_path: str, **kwargs):
        pass

    def save(self, model_path: str, **kwargs):
        pass

    def train(
            self,
            trn_data: Any,
            dev_data: Any,
            model_path: str,
            **kwargs) -> float:
        pass

    @abc.abstractmethod
    def decode(self, data: Any, **kwargs):
        pass

    def evaluate(self, data: Any, **kwargs):
        pass


class EnglishSegmenter(Segmenter):
    def decode(self, input_data, offsets=0, **kwargs):
        def sentence(sid, begin, end):
            return {
                SEN_ID: sid,
                TOK: input_data[begin:end],
                OFF: offsets[begin:end]
            }

        sentences = []
        begin = 0
        sid = 0
        right_quote = True

        for i, token in enumerate(input_data):
            t = token[0]
            if t == '"':
                right_quote = not right_quote

            if begin == i:
                if sentences and (is_right_bracket(t) or t ==
                                  u'\u201D' or t == '"' and right_quote):
                    d = sentences[-1]
                    d[TOK].append(token)
                    d[OFF].append(offsets[i])
                    begin = i + 1
            elif all(is_final_mark(c) for c in token):
                sentences.append(sentence(sid, begin, i + 1))
                sid += 1
                begin = i + 1

        if begin < len(input_data):
            sentences.append(sentence(sid, begin, len(input_data)))
            sid += 1

        return sentences



# ========================================================================
# Copyright 2018 Emory University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
import abc
import inspect
from typing import Optional, Tuple, Union

import numpy as np

from elit.util.structure import Document

__author__ = 'Jinho D. Choi'


class NLPState(abc.ABC):
    """
    :class:`NLPState` is an abstract class that takes a document and
    generates states to make predictions for an NLP task defined in its subclass.

    Abstract methods to be implemented:
      - :meth:`NLPState.init`
      - :meth:`NLPState.process`
      - :meth:`NLPState.has_next`
      - :meth:`NLPState.x`
      - :meth:`NLPState.y`
    """

    def __init__(self, document: Document, key: str):
        """
        :param document: the input document.
        :param key: the key to the input document where the predicted labels are to be saved.
        """
        self.document = document
        self.key = key

    def __iter__(self) -> 'NLPState':
        self.init()
        return self

    def __next__(self) -> Tuple[np.ndarray, Optional[int]]:
        if not self.has_next():
            raise StopIteration
        x, y = self.x, self.y
        self.process()
        return x, y

    @abc.abstractmethod
    def init(self, **kwargs):
        """
        Initializes the state of the input document.
        """
        raise NotImplementedError(
            '%s.%s()' %
            (self.__class__.__name__, inspect.stack()[0][3]))

    @abc.abstractmethod
    def process(self, *args):
        """
        :param args: parameters to be applied to the current state.
        Applies the parameters to the current state, then processes onto the next state.
        """
        raise NotImplementedError(
            '%s.%s()' %
            (self.__class__.__name__, inspect.stack()[0][3]))

    @abc.abstractmethod
    def has_next(self) -> bool:
        """
        :return: ``True`` if there exists the next state to be processed; otherwise, ``False``.
        """
        raise NotImplementedError(
            '%s.%s()' %
            (self.__class__.__name__, inspect.stack()[0][3]))

    @property
    def x(self) -> Union[np.ndarray, Tuple[np.ndarray, ...]]:
        """
        :return: the feature vector or the tuple of feature vectors extracted from the current state.
        :rtype: numpy.ndarray or tuple(numpy.ndarray, ....)
        """
        raise NotImplementedError(
            '%s.%s()' %
            (self.__class__.__name__, inspect.stack()[0][3]))

    @property
    @abc.abstractmethod
    def y(self) -> Optional[Union[int, Tuple[int, ...]]]:
        """
        :return: the class ID of the gold label for the current state if available; otherwise ``None``.
        :rtype: int or tuple(int, ....) or None
        """
        raise NotImplementedError(
            '%s.%s()' %
            (self.__class__.__name__, inspect.stack()[0][3]))



# ========================================================================
# Copyright 2018 Emory University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
from typing import List, Optional

import numpy as np
from elit.util.structure import Document, to_gold, to_out
from elit.util.vsm import LabelMap, VectorSpaceModel

__author__ = 'Jinho D. Choi'


# ======================================== State =========================

class DocumentClassifierState(BatchState):
    """
    DocumentClassificationBatchState labels input document with a certain class in batch mode.
    """

    def __init__(self,
                 document: Document,
                 key: str,
                 sentence_level: bool,
                 label_map: LabelMap,
                 word_vsm: VectorSpaceModel,
                 maxlen: int):
        """
        :param document: an input document.
        :param key: the key to the input document where the predicted labels are saved.
        :param sentence_level: if True, treats each sentence as a document.
        :param label_map: collects class labels during training and maps them to unique IDs.
        :param word_vsm: a vector space model for word embeddings.
        :param maxlen: the maximum length of a document;
                       for sentence-level, it is set to the maximum number of words in a sentence, and
                       for document-level, it is set to the maximum number of words across all sentences.
        """
        super().__init__(document, key)
        self.sentence_level = sentence_level
        self.label_map = label_map

        # initialize gold-standard labels and embeddings
        key_gold = to_gold(key)

        if self.sentence_level:
            self.gold = [
                s[key_gold] for s in document] if key_gold in document.sentences[0] else None
            self.embs = [
                word_vsm.embedding_matrix(
                    s.tokens,
                    maxlen) for s in document]
        else:
            self.gold = [document[key_gold]] if key_gold in document else None
            self.emb = [word_vsm.embedding_matrix(document.tokens, maxlen)]

        # self.init()
        self.doc_id = 0

    def init(self):
        """
        Initializes the pointers to the first document.
        """
        self.doc_id = 0

    def process(self):
        """
        Processes to the next document.
        """
        self.doc_id += 1

    @property
    def has_next(self) -> bool:
        """
        :return: False if no more document is left to be classified; otherwise, True.
        """
        if self.sentence_level:
            return self.doc_id < len(self.document)
        return self.doc_id == 0

    @property
    def x(self) -> List[np.ndarray]:
        """
        :return: the embedding matrix of the current document.
        """
        return self.embs[self.doc_id]

    @property
    def y(self) -> Optional[int]:
        """
        :return: the class ID of the current document's gold-standard label if available; otherwise, None.
        """
        return self.label_map.add(
            self.gold[self.doc_id]) if self.gold is not None else None

    def assign(self, output: np.ndarray, begin: int = 0) -> int:
        """
        Assigns the predicted output to the input document.
        :param output: a matrix where each row contains prediction scores of the corresponding document.
        :param begin: the row index of the output matrix corresponding to this document.
        :return: the row index of the output matrix to be assigned by the next document.
        """
        key_out = to_out(self.key)

        if self.sentence_level:
            for i, sentence in enumerate(self.document):
                sentence[key_out] = output[begin + i]
            return begin + len(self.document)
        else:
            self.document[key_out] = output
            return begin + 1



# ========================================================================
# Copyright 2018 Emory University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
import itertools
from typing import List, Tuple

__author__ = 'Jinho D. Choi'

DOC_ID = 'doc_id'   # document ID
SEN = 'sen'         # sentences
SEN_ID = 'sen_id'   # sentence ID
TOK = 'tok'
OFF = 'off'
LEM = 'lem'
POS = 'pos'
NER = 'ner'
DEP = 'dep'
COREF = 'coref'
SENTI = 'senti'


# sentiment analysis labels
SENTI_POSITIVE = 'p'
SENTI_NEGATIVE = 'n'
SENTI_NEUTRAL = '0'


class Sentence(dict):
    def __init__(self, d=None, **kwargs):
        """
        :param d: a dictionary, if not None, all of whose fields are added to this sentence.
        :type d: dict
        :param kwargs: additional fields to be added; if keys already exist; the values are overwritten with these.
        """
        super().__init__()
        self._iter = -1

        if d is not None: self.update(d)
        self.update(kwargs)
        self._tokens = self.setdefault(TOK, [])

    def __len__(self):
        """
        :return: the number of tokens in the sentence.
        """
        return len(self.tokens)

    def __iter__(self):
        self._iter = -1
        return self

    def __next__(self):
        self._iter += 1
        if self._iter >= len(self.tokens):
            raise StopIteration
        return self._tokens[self._iter]

    @property
    def tokens(self):
        """
        :return: the list of tokens in the sentence.
        :rtype: list of str
        """
        return self._tokens

    @property
    def part_of_speech_tags(self):
        """
        :return: the list of part-of-speech tags corresponding to the tokens in this sentence.
        :rtype: list of str
        """
        return self[POS]


class Document(dict):
    def __init__(self, d=None, **kwargs):
        """
        :param d: a dictionary, if not None, all of whose fields are added to this document.
        :type d: dict
        :param kwargs: additional fields to be added; if keys already exist; the values are overwritten with these.
        """
        super().__init__()
        self._iter = -1

        if d is not None: self.update(d)
        self.update(kwargs)
        self._sentences = self.setdefault(SEN, [])

    def __len__(self):
        """
        :return: the number of sentences in the document.
        :rtype: int
        """
        return len(self.sentences)

    def __iter__(self):
        self._iter = -1
        return self

    def __next__(self):
        self._iter += 1
        if self._iter >= len(self.sentences):
            raise StopIteration
        return self._sentences[self._iter]

    @property
    def sentences(self) -> List[Sentence]:
        """
        :return: the list of sentences in the document.
        :rtype: list of Sentence
        """
        return self._sentences

    @property
    def tokens(self):
        """
        :return: list of tokens across all sentences.
        :rtype: list of str
        """
        tokens = []
        for sentence in self.sentences:
            tokens.extend(sentence.tokens)
        return tokens


class Node(object):
    """
    Node is used for constructing the NLP structure.

    """
    def __init__(self):
        self._parent = None
        self._left_sibling = None
        self._right_sibling = None
        self._children = []

    @property
    def parent(self):
        """

        :return: parent of current node
        :rtype: Node
        """
        return self._parent

    @parent.setter
    def parent(self, node):
        if not isinstance(node, Node):
            raise ValueError('parent must be an Node')
        if self.has_parent():
            # not sure what's going on here?
            self.parent.remove_child(self)
        else:
            node.add_child(self)

    @property
    def left_sibling(self):
        """

        :return: left libling of current node
        :rtype: Node
        """
        return self._left_sibling

    @left_sibling.setter
    def left_sibling(self, node):
        if not isinstance(node, Node):
            raise ValueError('sibling must be an Node')
        if node is not None:
            self._left_sibling = node

    @property
    def right_sibling(self):
        """

        :return: right libling of current node
        :rtype: Node
        """
        return self.right_sibling

    @right_sibling.setter
    def right_sibling(self, node):
        if not isinstance(node, Node):
            raise ValueError('sibling must be an Node')
        if node is not None:
            self._right_sibling = node

    @property
    def children(self):
        """
        the list of children

        :return: the list of children
        :rtype: list
        """
        return self._children

    def index_of(self, child):
        """
        the index of the child

        :type child: Node
        :param child: a child node
        :return: the index of the child node. return None if the given node is not a child
        :rtype: int or None
        """
        try:
            return self.children.index(child)
        except ValueError:
            return None

    # Child

    def child(self, index):
        """


        :type index: int
        :param index: index of a child node
        :return: the index'th child of this node if exists; otherwise, None
        :rtype: Node or None
        """
        try:
            return self.children[index]
        except IndexError:
            return None

    def first_child(self):
        """

        :return: the first child of this node if exists; otherwise, None
        :rtype: Node or None
        """
        # return self.child(0)
        return self.find_child_by(index=0)

    def last_child(self):
        """

        :return: the last child of this node if exists; otherwise, None
        :rtype: Node or None
        """
        # return self.child(-1)
        return self.find_child_by(index=-1)

    def find_child_by(self, matcher=lambda n: n, index=0):
        """

        :type matcher: lambda
        :type index: int
        :param matcher: the condition
        :param index: displacement (0: 1st, 1: 2nd, etc.)
        :return: the order'th child matching the specific condition of this node if exists; otherwise, None
        :rtype: Node or None
        """
        return self.find_children_by(matcher)[index]

    def add_child(self, node, index=None):
        """
        Adds a node as the index'th child of this node if it is not already a child of this node.

        :type node: Node
        :type index: int
        :param node: node is added to children
        :param index: add child to this position
        :return: if the node is added successfully, retun True; otherwise, False
        :rtype: bool
        """
        if index is None:
            index = self.index_of(node)
        if self.is_parent_of(node):
            return False
        else:
            if node.has_parent():
                node.parent.remove_child(node)
            node.parent = self
            self.children.insert(index, node)
            self._siblings(self.child(index - 1), node)
            self._siblings(node, self.child(index + 1))
            return True

    def set_child(self, node, index):
        """
        Sets a node as the index'th child of this node if it is not already a child of this node.

        :type node: Node
        :type index: int
        :param node: node is set at the index position
        :param index: set a child to this position
        :return: the previously index'th node if added; otherwise, None
        :rtype: Node or None
        """
        if not self.is_parent_of(node):
            if node.has_parent():
                node.parent.remove_child(node)
            node.parent = self
            old_node = self.children[index]
            self.children[index] = node
            self._siblings(self.child(index - 1), node)
            self._siblings(node, self.child(index + 1))
            old_node.isolate()
            return old_node
        else:
            return None

    def remove_child(self, node):
        """
        Removes a child from this node.

        :type node: Node
        :param node: remove node
        :return: True if the node exist; otherwise, False
        :rtype: bool
        """
        index = self.index_of(node)
        if index is not None:
            self._siblings(self.child(index - 1), self.child(index + 1))
            self.children.remove(node)  #
            node.isolate()  # is isolation necessary ?
            return True
        else:
            return False

    def replace_child_from(self, old_node, new_node):
        """
        Replaces the child from old_node to new_node

        :type old_node: Node
        :type new_node: Node
        :param old_node: old child
        :param new_node: new child
        :return: True if the child is replaced successfully; otherwise, False
        :rtype: bool
        """

        index = self.index_of(old_node)
        if index is not None:
            self.children[index] = new_node
            self._siblings(self.child(index - 1), new_node)
            self._siblings(new_node, self.child(index + 1))
            old_node.isolate()
            return True
        else:
            return False

    def remove(self):
        """
        Removes this node from its parent.
        If this is the only child, removeToken its parent from its grandparent, and applies
        this logic recursively to the ancestors.

        """
        node = self
        while node.has_parent():
            parent = node.parent
            parent.remove_child(node)
            if parent.has_child():
                break
            node = parent

    def has_child(self):
        """
        Whether current node has any child or not

        :return: True if this node has any child; otherwise, False
        :rtype: bool
        """
        return True if self.children else False

    def is_child_of(self, node):
        """
        Whether current node is a child of the node or not

        :param node: Node
        :param node: Node
        :return: True if this node is a child of the specific node; otherwise, False
        :rtype: bool
        """
        return node is not None and self.parent is node

    def contains_child(self, matcher=lambda n: n):
        """
        Whether current node contains a child with the matcher or not

        :type matcher: lambda
        :param matcher: The matcher of condition
        :return: True if this node contains a child of the matcher; otherwise, False
        :rtype: bool
        """
        return True if list(filter(matcher, self.children)) else False

    # Descendants

    def num_of_children(self):
        """

        :return: numbers of children
        :rtype: int
        """
        return len(self.children)

    def find_children_in(self, first_child_index=0, last_child_index=None):
        """
        The sublist begins at the specific position and extends to the end.

        :type first_child_index: int
        :type last_child_index: int
        :param first_child_index: the ID of the first child (inclusive)
        :param last_child_index: the ID of the last child (exclusive)
        :return: an immutable list of sub-children
        :rtype: list
        """
        return self.children[first_child_index:last_child_index]

    def find_children_by(self, matcher=lambda n: n):
        """

        :type matcher: lambda
        :param matcher: the condition
        :return: the list of children matching the specific condition.
        :rtype: list
        """
        return list(filter(matcher, self.children))

    def grand_children(self):
        """

        :return: the list of grand-children
        :rtype: list
        """
        return self.second_order('children')

    def find_first_descendant_by(self, matcher=lambda n: True):
        """

        :type matcher: lambda
        :param matcher: the condition
        :return: the first descendant matching the specific condition
        :rtype: Node or None
        """
        for node in self.children:
            if matcher(node):
                return node

            node = self.find_first_descendant_by(node.children, matcher)
            if node is not None:
                return node
        return None

    def find_first_lowest_chained_descendant_by(self, matcher):
        """

        :type matcher: lambda
        :param matcher: the condition
        :return: the first lowest descendant whose intermediate ancestors to this node all match the specific condition
        :rtype: Node or None
        """
        node = self.find_child_by(matcher)
        descendant = None

        while node is not None:
            descendant = node
            node = node.find_child_by(matcher)

        return descendant

    def descendants(self):
        """

        :return: the list of all descendents (excluding this node).
        :rtype: list
        """
        result = []
        if not self.has_child():
            return [self]
        else:
            for child in self.children:
                result = result + child.descendants
            return result

    def find_descendants_in(self, depth):
        """

        :type depth: int
        :param depth: the level of the descendents to be retrieved (1: children, 2: childreNLPNode + grand-children, etc.).
        :return: the list of descendents (excluding this node).
        :rtype: list
        """
        result = []
        if depth <= 0:
            return result
        else:
            for child in self.children:
                result = result + child.find_descendants_in(depth - 1)
            return result

    def find_single_chained_by(self, matcher=lambda n: True):
        """

        :type matcher: lambda
        :param matcher: the condition
        :return: a node in this node's subtree (including self) matching the specific condition if it is single-chained.
        :rtype: Node or None
        """
        node = self
        while node is not None:
            if matcher(node):
                return node
            if node.num_of_children() is 1:
                node = node.first_child()
            else:
                break
        return None

    def adapt_descendants_from(self, node):
        """

        :type node: Node
        :param node:
        """
        for child in node.children:
            child.parent = self

    def is_descendants_of(self, node):
        """

        :type node: Node
        :param node: the node
        :return: True if the node is a descendant of the specific node
        :rtype: bool
        """
        return self.find_nearest_node_by(
            getter='parent', matcher=lambda n: n is node) is not None

    # Ancestors

    def grand_parent(self):
        """

        :return: the grandparent of this node if exists; otherwise, None
        :rtype: Node or None
        """
        return self.ancestor(height=2)

    def ancestor(self, height):
        """

        :type height: int
        :param height:  height of the ancestor from this node (1: parent, 2: grandparent, etc.).
        :return: the lowest ancestor matching the specific condition
        :rtype: Node
        """
        return self.find_node_in(order=height, getter='parent')

    def find_lowest_ancestor_by(self, matcher=lambda n: True):
        """

        :type matcher: lambda
        :param matcher: the condition
        :return: the lowest ancestor matching the specific condition
        :rtype: Node
        """
        return self.find_nearest_node_by('parent', matcher)

    def find_highest_ancestor_by(self, matcher=lambda n: True):
        """

        :type matcher: lambda
        :param matcher: the condition
        :return: the highest ancestor where all the intermediate ancestors match the specific condition
        :rtype: Node
        """
        node = self.parent
        ancestor = None
        while node is not None:
            if matcher(node):
                ancestor = node
            else:
                break
            node = node.parent
        return ancestor

    def lowest_common_ancestor(self, node):
        """

        :type node: Node
        :param node: the node
        :return: the lowest common ancestor of this node and the specified node
        :rtype: Node or None
        """
        ancestor_set = self.ancestor_set()
        ancestor_set.add(self)

        while node is not None:
            if node in ancestor_set:
                return node
            node = node.parent

        return None

    def ancestor_set(self):
        """

        :return: the setToken of all ancestors of this node
        :rtype: set
        """
        ancestor_set = set()
        node = self.parent
        while node is not Node:
            ancestor_set.add(node)
            node = node.parent
        return ancestor_set

    def is_parent_of(self, node):
        """

        :type node: Node
        :param node: the node
        :return: True if this node is the parent of the specific node; otherwise, False
        :rtype: bool
        """
        return node.is_child_of(self)

    def is_ancestor_of(self, node):
        """

        :type node: Node
        :param node: the node
        :return: True if the node is a descendant of the specific node; otherwise, False
        :rtype: bool
        """
        return node.is_descendants_of(self)

    def has_parent(self, matcher=lambda n: True):
        """

        :param matcher: the condition
        :return: True if this node has a parent; otherwise, False
        :rtype: bool
        """
        return self.parent is not None and matcher(self)

    def has_grand_parent(self):
        """

        :return: True if this node has a grand parent; otherwise, False
        :rtype: bool
        """
        return self.grand_parent() is not None

    # Siblings

    def siblings(self):
        """

        :return: the list of all siblings
        :rtype: list
        """
        return list(
            filter(
                lambda child: child is not self,
                self.parent.children)) if self.has_parent() else []

    def left_nearest_sibling(self, order=0):
        """

        :type order: int
        :param order: displacement (0: left-nearest, 1: 2nd left-nearest, etc.)
        :return: the order'th left-nearest sibling of this node if exists; otherwise, None
        :rtype: Node or None
        """
        return self.find_node_in(
            order + 1, 'left_sibling') if order >= 0 else None

    def find_left_nearest_sibling_by(self, matcher=lambda x: True):
        """

        :type matcher: lambda
        :param matcher: the condition
        :return: the left nearest sibling that matches the specific condition
        :rtype: Node
        """
        return self.find_nearest_node_by('left_sibling', matcher)

    def right_nearest_sibling(self, order=0):
        """

        :type order: int
        :param order: displacement (0: left-nearest, 1: 2nd left-nearest, etc.)
        :return: the order'th right-nearest sibling of this node if exists; otherwise, None
        :rtype: Node or None
        """
        return self.find_node_in(
            order + 1, 'right_sibling') if order >= 0 else None

    def find_right_nearest_sibling_by(self, matcher=lambda x: True):
        """

        :type matcher: lambda
        :param matcher: the condition
        :return: the right nearest sibling that matches the specific condition
        :rtype: Node
        """
        return self.find_nearest_node_by('right_sibling', matcher)

    def has_left_sibling(self, matcher=lambda x: True):
        """

        :type matcher: lambda
        :param matcher: the condition
        :return: Trur if this node has a left sibling; otherwise, False.
        """
        return self.left_sibling is not None or self.find_left_nearest_sibling_by(
            matcher)

    def has_right_sibling(self, matcher=lambda x: True):
        """

        :type matcher: lambda
        :param matcher: the condition
        :return: Trur if this node has a right sibling; otherwise, False.
        """
        return self.right_sibling is not None or self.find_right_nearest_sibling_by(
            matcher)

    def is_sibling_of(self, node):
        """

        :type node: Node
        :param node: the node
        :return: True if this node is a sibling of the specific node.
        :rtype: bool
        """
        return node.is_child_of(self.parent)

    def is_left_sibling_of(self, node):
        """

        :type node: Node
        :param node:
        :return: True if this node is a left sibling of the specific node
        :rtype: bool
        """
        return node is not None and self.parent is node.parent and self.find_nearest_node_by(
            'right_sibling', lambda n: n is node) is not None

    def is_right_sibling_of(self, node):
        """

        :type node: Node
        :param node:
        :return: True if this node is a right sibling of the specific node
        :rtype: bool
        """
        return node.is_left_sibling_of(self)

    # Helpers

    def find_node_in(self, order, getter):
        """

        :type order: int
        :type getter: str
        :param order: 0: self, 1: nearest, 2: second nearest, etc.
        :param getter: takes a node and returns a node.
        :return: the order'th node with respect to the getter.
        :rtype: Node
        """
        node = self
        for _ in range(order):
            if node is None:
                return None
            node = getattr(node, getter)
        return node

    def find_nearest_node_by(self, getter, matcher=lambda n: True):
        """

        :type getter: str
        :type matcher: lambda
        :param getter: takes a node and returns a node.
        :param matcher: takes a node and the supplement, and returns true if its field matches to the specific predicate.
        :return: the first node matching the specific condition
        :rtype: Node or None
        """
        node = getattr(self, getter)

        while node is not None:
            if matcher(node):
                return node
            node = getattr(node, getter)
        return None

    def isolate(self):
        """
        Isolates this node from its parent, children, and siblings
        """
        self._parent = None
        self._left_sibling = None
        self._right_sibling = None

    def second_order(self, getter):
        """

        :type getter: str
        :param getter: takes a node and returns a list of nodes.
        :return: the list of second order elements according to the getter.
        :rtype: list
        """
        return list(
            itertools.chain.from_iterable(
                list(
                    filter(
                        lambda n: getattr(
                            n, getter) if n is not self else None, getattr(
                            self, getter)))))

    def _siblings(self, left_node, right_node):
        """
        Sets two nodes siblings of each other.

        :type left_node: Node
        :type right_node: Node
        :param left_node:
        :param right_node:
        """
        right_node.left_sibling = left_node
        left_node.right_sibling = right_node


def to_gold(key: str) -> str:
    return key + '-gold'


def to_out(key: str) -> str:
    return key + '-out'


class BILOU:
    B = 'B'  # beginning
    I = 'I'  # inside
    L = 'L'  # last
    O = 'O'  # outside
    U = 'U'  # unit

    @classmethod
    def to_chunks(cls, tags: List[str], fix: bool = False) -> List[Tuple[int, int, str]]:
        """
        :param tags: a list of tags encoded by BILOU.
        :param fix: if True, fixes potential mismatches in BILOU (see :meth:`heuristic_fix`).
        :return: a list of tuples where each tuple contains (begin index (inclusive), end index (exclusive), label) of the chunk.
        """
        if fix: cls.heuristic_fix(tags)
        chunks = []
        begin = -1

        for i, tag in enumerate(tags):
            t = tag[0]

            if t == cls.B:
                begin = i
            elif t == cls.I:
                pass
            elif t == cls.L:
                if begin >= 0: chunks.append((begin, i + 1, tag[2:]))
                begin = -1
            elif t == cls.O:
                begin = -1
            elif t == cls.U:
                chunks.append((i, i + 1, tag[2:]))
                begin = -1

        return chunks

    @classmethod
    def heuristic_fix(cls, tags):
        """
        Use heuristics to fix potential mismatches in BILOU.
        :param tags: a list of tags encoded by BLIOU.
        """

        def fix(i, pt, ct, t1, t2):
            if pt == ct:
                tags[i][0] = t1
            else:
                tags[i - 1][0] = t2

        def aux(i):
            p = tags[i - 1][0]
            c = tags[i][0]
            pt = tags[i - 1][1:]
            ct = tags[i][1:]

            if p == cls.B:
                if c == cls.B:
                    fix(i, pt, ct, cls.I, cls.U)  # BB -> BI or UB
                elif c == cls.U:
                    fix(i, pt, ct, cls.L, cls.U)  # BU -> BL or UU
                elif c == cls.O:
                    tags[i - 1][0] = cls.U  # BO -> UO
            elif p == cls.I:
                if c == cls.B:
                    fix(i, pt, ct, cls.I, cls.L)  # IB -> II or LB
                elif c == cls.U:
                    fix(i, pt, ct, cls.I, cls.L)  # IU -> II or LU
                elif c == cls.O:
                    tags[i - 1][0] = cls.L  # IO -> LO
            elif p == cls.L:
                if c == cls.I:
                    fix(i, pt, ct, cls.I, cls.B)  # LI -> II or LB
                elif c == cls.L:
                    fix(i, pt, ct, cls.I, cls.B)  # LL -> IL or LB
            elif p == cls.O:
                if c == cls.I:
                    tags[i][0] = cls.B  # OI -> OB
                elif c == cls.L:
                    tags[i][0] = cls.B  # OL -> OB
            elif p == cls.U:
                if c == cls.I:
                    fix(i, pt, ct, cls.B, cls.B)  # UI -> BI or UB
                elif c == cls.L:
                    fix(i, pt, ct, cls.B, cls.B)  # UL -> BL or UB

        for idx in range(1, len(tags)): aux(idx)
        prev = tags[-1][0]

        if prev == cls.B:
            tags[-1][0] = cls.U
        elif prev == cls.I:
            tags[-1][0] = cls.L





# ========================================================================
# Copyright 2018 Emory University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
import argparse
import json
import logging
import pickle
from types import SimpleNamespace
from typing import Tuple, Optional, Union, Sequence, List

import mxnet as mx
import numpy as np
from mxnet.ndarray import NDArray

from elit.cli import ComponentCLI, set_logger, args_dict_str_float, args_tuple_int, args_reader, args_vsm, args_hidden, args_context, args_loss, args_ngram_conv, \
    args_fuse_conv
from elit.component import MXNetComponent
from elit.eval import Accuracy, F1
from elit.model import FFNNModel, NLPModel
from elit.state import NLPState
from elit.util.io import pkl, gln, group_states
from elit.util.iterator import SequenceIterator, BatchIterator
from elit.util.structure import Document, to_gold, BILOU
from elit.util.vsm import LabelMap, x_extract, Position2Vec

__author__ = 'Jinho D. Choi'


# ======================================== State =========================

class TokenTaggerState(NLPState):
    """
    :class:`TokenTaggingState` generates a state per token using the one-pass left-to-right decoding strategy
    and predicts tags for individual tokens.
    """

    def __init__(self,
                 document: Document,
                 key: str,
                 label_map: LabelMap,
                 vsm_list: Sequence[SimpleNamespace],
                 feature_windows: Tuple[int, ...],
                 padout: np.ndarray = None):
        """
        :param document: the input document.
        :param key: the key to each sentence in the document where predicted tags are to be saved.
        :param label_map: collects labels during training and maps them to unique class IDs.
        :param vsm_list: the sequence of namespace(:class:`elit.vsm.VectorSpaceModel`, key).
        :param feature_windows: the contextual windows for feature extraction.
        :param padout: the zero-vector whose dimension is the number of classes; if not ``None``, label embeddings are used as features.
        """
        super().__init__(document, key)
        self.label_map = label_map
        self.feature_windows = feature_windows
        self.padout = padout

        # initialize gold-standard labels if available
        key_gold = to_gold(key)
        self.gold = [s[key_gold]
                     for s in document] if key_gold in document.sentences[0] else None

        # initialize output and predicted tags
        self.output = []
        self.pred = []

        # initialize embeddings
        self.embs = [
            (vsm.model.sentence_embedding_list(
                document,
                vsm.key),
                vsm.model.pad) for vsm in vsm_list]
        if padout is not None:
            self.embs.append((self.output, padout))

        # the followings are initialized in self.init()
        self.sen_id = 0
        self.tok_id = 0
        self.init()

    def init(self):
        del self.output[:]
        del self.pred[:]
        self.sen_id = 0
        self.tok_id = 0

        for s in self.document:
            self.output.append([self.padout] * len(s))
            self.pred.append([None] * len(s))
            s[self.key] = self.pred[-1]

    def process(self, output: Optional[NDArray] = None):
        """
        :param output: the predicted output of the current token.

        Assigns the predicted output to the current token, then processes to the next token.
        """
        # apply the output to the current token
        if output is not None:
            self.output[self.sen_id][self.tok_id] = output.asnumpy()
            self.pred[self.sen_id][self.tok_id] = self.label_map.argmax(output)

        # process to the next token
        self.tok_id += 1
        if self.tok_id == len(self.document.sentences[self.sen_id]):
            self.sen_id += 1
            self.tok_id = 0

    def has_next(self) -> bool:
        """
        :return: ``False`` if no more token is left to be tagged; otherwise, ``True``.
        """
        return 0 <= self.sen_id < len(self.document)

    @property
    def x(self) -> np.ndarray:
        """
        :return: the feature matrix of the current token.
        """
        l = ([x_extract(self.tok_id, w, emb[self.sen_id], pad)
              for w in self.feature_windows] for emb, pad in self.embs)
        return np.column_stack(l)

    @property
    def y(self) -> Optional[int]:
        """
        :return: the class ID of the current token's gold-standard label if available; otherwise, None.
        """
        return self.label_map.add(
            self.gold[self.sen_id][self.tok_id]) if self.gold is not None else None


# ======================================== EvalMetric ====================

class TokenAccuracy(Accuracy):
    def __init__(self, key: str):
        super().__init__()
        self.key = key

    def update(self, document: Document):
        for sentence in document:
            gold = sentence[to_gold(self.key)]
            pred = sentence[self.key]
            self.correct += len([1 for g, p in zip(gold, pred) if g == p])
            self.total += len(sentence)


class ChunkF1(F1):
    def __init__(self, key: str):
        super().__init__()
        self.key = key

    def update(self, document: Document):
        for sentence in document:
            gold = BILOU.to_chunks(sentence[to_gold(self.key)])
            pred = BILOU.to_chunks(sentence[self.key])
            self.correct += len(set.intersection(set(gold), set(pred)))
            self.p_total += len(pred)
            self.r_total += len(gold)


# ======================================== Component =====================

class TokenTagger(MXNetComponent):
    """
    :class:`TokenTagger` provides an abstract class to implement a tagger that predicts a tag for every token.
    """

    def __init__(self,
                 ctx: mx.Context,
                 vsm_list: List[SimpleNamespace]):
        """
        :param ctx: the (list of) device context(s) for :class:`mxnet.gluon.Block`.
        :param vsm_list: the sequence of namespace(model, key),
                         where the key indicates the key of the values to retrieve embeddings for (e.g., tok).
        """
        super().__init__(ctx)
        self.vsm_list = vsm_list

        # to be loaded/saved
        self.key = None
        self.sequence = None
        self.chunking = None
        self.label_map = None
        self.feature_windows = None
        self.position_embedding = None
        self.padout = None
        self.input_config = None
        self.output_config = None
        self.fuse_conv_config = None
        self.ngram_conv_config = None
        self.hidden_configs = None

    def __str__(self):
        s = ('Token Tagger',
             '- type: %s' % ('chunking' if self.chunking else 'tagging'),
             '- sequence: %r' % self.sequence,
             '- feature windows: %s' % str(self.feature_windows),
             '- label embedding: %r' % (self.padout is not None),
             '- position embedding: %r' % self.position_embedding,
             '- %s' % str(self.model))
        return '\n'.join(s)

    # override
    def init(self,
             key: str,
             sequence: bool,
             chunking: bool,
             num_class: int,
             feature_windows: Tuple[int, ...],
             position_embedding: bool,
             label_embedding: bool,
             input_dropout: float = 0.0,
             fuse_conv_config: Optional[SimpleNamespace] = None,
             ngram_conv_config: Optional[SimpleNamespace] = None,
             hidden_configs: Optional[Tuple[SimpleNamespace]] = None,
             initializer: mx.init.Initializer = None,
             **kwargs):
        """
        :param key: the key to each sentence where the tags are to be saved.
        :param sequence: if ``True``, run in sequence mode; otherwise, batch mode.
        :param chunking: if ``True``, run chunking instead of tagging.
        :param num_class: the number of classes (part-of-speech tags).
        :param feature_windows: the content windows for feature extraction.
        :param position_embedding: if ``True``, use position embeddings as features.
        :param label_embedding: if ``True``, use label embeddings as features.
        :param input_dropout: the dropout rate to be applied to the input layer.
        :param fuse_conv_config: the configuration for the fuse convolution layer.
        :param ngram_conv_config: the configuration for the n-gram convolution layer.
        :param hidden_configs: the configurations for the hidden layers.
        :param initializer: the weight initializer for :class:`mxnet.gluon.Block`.
        :param kwargs: extra parameters to initialize :class:`mxnet.gluon.Block`.
        """
        # configuration
        if position_embedding:
            self.vsm_list.append(
                SimpleNamespace(
                    model=Position2Vec(),
                    key=None))

        input_dim = sum([vsm.model.dim for vsm in self.vsm_list])
        if label_embedding:
            input_dim += num_class
        input_config = NLPModel.namespace_input_layer(
            row=len(feature_windows), col=input_dim, dropout=input_dropout)
        output_config = NLPModel.namespace_output_layer(num_class)
        if initializer is None:
            initializer = mx.init.Xavier(magnitude=2.24, rnd_type='gaussian')

        # initialization
        self.key = key
        self.sequence = sequence
        self.chunking = chunking
        self.label_map = LabelMap()
        self.feature_windows = feature_windows
        self.position_embedding = position_embedding
        self.padout = np.zeros(num_class).astype(
            'float32') if label_embedding else None
        self.input_config = input_config
        self.output_config = output_config
        self.fuse_conv_config = fuse_conv_config
        self.ngram_conv_config = ngram_conv_config
        self.hidden_configs = hidden_configs

        self.model = FFNNModel(
            self.input_config,
            self.output_config,
            self.fuse_conv_config,
            self.ngram_conv_config,
            self.hidden_configs,
            **kwargs)
        self.model.collect_params().initialize(initializer, ctx=self.ctx)
        logging.info(self.__str__())

    # override
    def load(self, model_path: str, **kwargs):
        """
        :param model_path: the path to a pre-trained model to be loaded.
        :type model_path: str
        """
        with open(pkl(model_path), 'rb') as fin:
            self.key = pickle.load(fin)
            self.sequence = pickle.load(fin)
            self.chunking = pickle.load(fin)
            self.label_map = pickle.load(fin)
            self.feature_windows = pickle.load(fin)
            self.position_embedding = pickle.load(fin)
            self.padout = pickle.load(fin)
            self.input_config = pickle.load(fin)
            self.output_config = pickle.load(fin)
            self.fuse_conv_config = pickle.load(fin)
            self.ngram_conv_config = pickle.load(fin)
            self.hidden_configs = pickle.load(fin)

        self.model = FFNNModel(
            self.input_config,
            self.output_config,
            self.fuse_conv_config,
            self.ngram_conv_config,
            self.hidden_configs,
            **kwargs)
        self.model.collect_params().load(gln(model_path), self.ctx)
        if self.position_embedding:
            self.vsm_list.append(
                SimpleNamespace(
                    model=Position2Vec(),
                    key=None))
        logging.info(self.__str__())

    # override
    def save(self, model_path, **kwargs):
        """
        :param model_path: the filepath where the model is to be saved.
        :type model_path: str
        """
        with open(pkl(model_path), 'wb') as fout:
            pickle.dump(self.key, fout)
            pickle.dump(self.sequence, fout)
            pickle.dump(self.chunking, fout)
            pickle.dump(self.label_map, fout)
            pickle.dump(self.feature_windows, fout)
            pickle.dump(self.position_embedding, fout)
            pickle.dump(self.padout, fout)
            pickle.dump(self.input_config, fout)
            pickle.dump(self.output_config, fout)
            pickle.dump(self.fuse_conv_config, fout)
            pickle.dump(self.ngram_conv_config, fout)
            pickle.dump(self.hidden_configs, fout)

        self.model.collect_params().save(gln(model_path))

    # override
    def data_iterator(self,
                      documents: Sequence[Document],
                      batch_size: int,
                      shuffle: bool,
                      label: bool,
                      **kwargs) -> Union[BatchIterator,
                                         SequenceIterator]:
        if self.sequence:
            def create(d: Document) -> TokenTaggerState:
                return TokenTaggerState(
                    d,
                    self.key,
                    self.label_map,
                    self.vsm_list,
                    self.feature_windows,
                    self.padout)

            states = group_states(documents, create)
        else:
            states = [
                TokenTaggerState(
                    d,
                    self.key,
                    self.label_map,
                    self.vsm_list,
                    self.feature_windows) for d in documents]

        iterator = SequenceIterator if self.sequence else BatchIterator
        return iterator(states, batch_size, shuffle, label, **kwargs)

    # override
    def eval_metric(self) -> Union[TokenAccuracy, ChunkF1]:
        """
        :return: :class:`ChunkF1` if self.chunking else :class:`TokenAccuracy`.
        """
        return ChunkF1(self.key) if self.chunking else TokenAccuracy(self.key)


# ======================================== Command-Line Interface ========

class TokenTaggerCLI(ComponentCLI):
    def __init__(self):
        super().__init__('token_tagger', 'Token Tagger')

    # override
    @classmethod
    def train(cls, args):
        # create a arg-parser
        parser = argparse.ArgumentParser(
            description='Train a token tagger',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)

        # data
        group = parser.add_argument_group("data arguments")

        group.add_argument('-t', '--trn_path', type=str, metavar='filepath',
                           help='filepath to the training data (input)')
        group.add_argument('-d', '--dev_path', type=str, metavar='filepath',
                           help='filepath to the development data (input)')
        group.add_argument(
            '-m',
            '--model_path',
            type=str,
            metavar='filepath',
            default=None,
            help='filepath to the model data (output); if not set, the model is not saved')
        group.add_argument(
            '-r',
            '--reader',
            type=args_reader,
            metavar='json or tsv=(str:int)(,#1)*',
            default=args_reader('json'),
            help='type of the reader and its configuration to match the data format (default: json)')
        group.add_argument(
            '-v',
            '--vsm_list',
            type=args_vsm,
            metavar='(fasttext|word2vec:key:filepath)( #1)*',
            nargs='+',
            help='list of (type of vector space model, key, filepath)')
        group.add_argument(
            '-l',
            '--log',
            type=str,
            metavar='filepath',
            default=None,
            help='filepath to the logging file; if not set, use stdout')

        # tagger
        group = parser.add_argument_group("tagger arguments")

        group.add_argument(
            '-key',
            '--key',
            type=str,
            metavar='str',
            help='key to the document dictionary where the predicted tags are to be stored')
        group.add_argument(
            '-seq',
            '--sequence',
            action='store_true',
            help='if set, run in sequence mode; otherwise, batch mode')
        group.add_argument(
            '-chu',
            '--chunking',
            action='store_true',
            help='if set, tag chunks (e.g., named entities); otherwise, tag tokens (e.g., part-of-speech tags)')
        group.add_argument(
            '-fw',
            '--feature_windows',
            type=args_tuple_int,
            metavar='int(,int)*',
            default=args_tuple_int('3,2,1,0,-1,-2,-3'),
            help='content windows for feature extraction (default: 3,2,1,0,-1,-2,-3)')
        group.add_argument('-pe', '--position_embedding', action='store_true',
                           help='if set, use position embeddings as features')
        group.add_argument('-le', '--label_embedding', action='store_true',
                           help='if set, use label embeddings as features')

        # network
        group = parser.add_argument_group("network arguments")

        group.add_argument('-nc', '--num_class', type=int, metavar='int',
                           help='number of classes')
        group.add_argument(
            '-ir',
            '--input_dropout',
            type=float,
            metavar='float',
            default=0.0,
            help='dropout rate applied to the input layer (default: 0.0)')
        group.add_argument(
            '-fc',
            '--fuse_conv_config',
            type=args_fuse_conv,
            metavar='(filters:activation:dropout)',
            default=None,
            help='configuration for the fuse convolution layer (default: None)')
        group.add_argument(
            '-cc',
            '--ngram_conv_config',
            type=args_ngram_conv,
            metavar='(ngrams:filters:activation:pool:dropout)',
            default=args_ngram_conv('1,2,3,4,5:128:relu:none:0.2'),
            help='configuration for the n-gram convolution layer (default: 1,2,3,4,5:128:relu:none:0.2)')
        group.add_argument(
            '-hc',
            '--hidden_configs',
            type=args_hidden,
            metavar='(dim:activation:dropout)( #1)*',
            nargs='+',
            default=None,
            help='configuration for the hidden layers (default: None)')

        # training
        group = parser.add_argument_group("arguments for training")

        group.add_argument(
            '-cx',
            '--context',
            type=args_context,
            metavar='[cg]int',
            nargs='+',
            default=mx.cpu(),
            help='device context(s)')
        group.add_argument(
            '-ep',
            '--epoch',
            type=int,
            metavar='int',
            default=100,
            help='number of epochs')
        group.add_argument(
            '-tb',
            '--trn_batch',
            type=int,
            metavar='int',
            default=64,
            help='batch size for the training data')
        group.add_argument(
            '-db',
            '--dev_batch',
            type=int,
            metavar='int',
            default=2048,
            help='batch size for the development data')
        group.add_argument(
            '-lo',
            '--loss',
            type=args_loss,
            metavar='str',
            default=None,
            help='loss function')
        group.add_argument(
            '-op',
            '--optimizer',
            type=str,
            metavar='str',
            default='adagrad',
            help='optimizer')
        group.add_argument(
            '-opp',
            '--optimizer_params',
            type=args_dict_str_float,
            metavar='([A-Za-z0-9_-]+):(\\d+)',
            default=args_dict_str_float('learning_rate:0.01'),
            help='optimizer parameters')

        # arguments
        args = parser.parse_args(args)
        set_logger(args.log)

        args.vsm_list = [
            SimpleNamespace(
                model=n.type(
                    n.filepath),
                key=n.key) for n in args.vsm_list]
        if isinstance(args.context, list) and len(args.context) == 1:
            args.context = args.context[0]
        if args.loss is None:
            args.loss = mx.gluon.loss.SoftmaxCrossEntropyLoss()
        initializer = mx.init.Xavier(magnitude=2.24, rnd_type='gaussian')

        # component
        comp = TokenTagger(args.context, args.vsm_list)
        comp.init(
            args.key,
            args.sequence,
            args.chunking,
            args.num_class,
            args.feature_windows,
            args.position_embedding,
            args.label_embedding,
            args.input_dropout,
            args.fuse_conv_config,
            args.ngram_conv_config,
            args.hidden_configs,
            initializer)

        # data
        trn_docs = args.reader.type(
            args.trn_path, args.reader.params, args.key)
        dev_docs = args.reader.type(
            args.dev_path, args.reader.params, args.key)

        # train
        comp.train(
            trn_docs,
            dev_docs,
            args.model_path,
            args.trn_batch,
            args.dev_batch,
            args.epoch,
            args.loss,
            args.optimizer,
            args.optimizer_params)
        logging.info('# of classes: %d' % len(comp.label_map))

    # override
    @classmethod
    def decode(cls, args):
        # create a arg-parser
        parser = argparse.ArgumentParser(
            description='Decode with the token tagger',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)

        # data
        group = parser.add_argument_group("data arguments")

        group.add_argument('-i', '--input_path', type=str, metavar='filepath',
                           help='filepath to the input data')
        group.add_argument(
            '-o',
            '--output_path',
            type=str,
            metavar='filepath',
            default=None,
            help='filepath to the output data (default: input_path.json)')
        group.add_argument('-m', '--model_path', type=str, metavar='filepath',
                           default=None,
                           help='filepath to the model data')
        group.add_argument(
            '-r',
            '--reader',
            type=args_reader,
            metavar='json or tsv=(str:int)(,#1)*',
            default=args_reader('json'),
            help='type of the reader and its configuration to match the data format (default: json)')
        group.add_argument(
            '-v',
            '--vsm_list',
            type=args_vsm,
            metavar='(fasttext|word2vec:key:filepath)( #1)*',
            nargs='+',
            help='list of (type of vector space model, key, filepath)')
        group.add_argument(
            '-l',
            '--log',
            type=str,
            metavar='filepath',
            default=None,
            help='filepath to the logging file; if not set, use stdout')

        # evaluation
        group = parser.add_argument_group("arguments for decoding")

        group.add_argument(
            '-cx',
            '--context',
            type=args_context,
            metavar='[cg]int',
            nargs='+',
            default=mx.cpu(),
            help='device context(s)')
        group.add_argument(
            '-ib',
            '--input_batch',
            type=int,
            metavar='int',
            default=2048,
            help='batch size for the input data')

        # arguments
        args = parser.parse_args(args)
        set_logger(args.log)

        args.vsm_list = [
            SimpleNamespace(
                model=n.type(
                    n.filepath),
                key=n.key) for n in args.vsm_list]
        if isinstance(args.context, list) and len(args.context) == 1:
            args.context = args.context[0]
        if args.output_path is None:
            args.output_path = args.input_path + '.json'

        # component
        comp = TokenTagger(args.context, args.vsm_list)
        comp.load(args.model_path)

        # data
        docs = args.reader.type(args.input_path, args.reader.params)

        # evaluate
        comp.decode(docs, args.input_batch)

        with open(args.output_path, 'w') as fout:
            json.dump(docs, fout)

    # override
    @classmethod
    def evaluate(cls, args):
        # create a arg-parser
        parser = argparse.ArgumentParser(
            description='Evaluate the token tagger',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)

        # data
        group = parser.add_argument_group("data arguments")

        group.add_argument('-d', '--dev_path', type=str, metavar='filepath',
                           help='filepath to the development data (input)')
        group.add_argument('-m', '--model_path', type=str, metavar='filepath',
                           default=None,
                           help='filepath to the model data')
        group.add_argument(
            '-r',
            '--reader',
            type=args_reader,
            metavar='json or tsv=(str:int)(,#1)*',
            default=args_reader('json'),
            help='type of the reader and its configuration to match the data format (default: json)')
        group.add_argument(
            '-v',
            '--vsm_list',
            type=args_vsm,
            metavar='(fasttext|word2vec:key:filepath)( #1)*',
            nargs='+',
            help='list of (type of vector space model, key, filepath)')
        group.add_argument(
            '-l',
            '--log',
            type=str,
            metavar='filepath',
            default=None,
            help='filepath to the logging file; if not set, use stdout')

        # evaluation
        group = parser.add_argument_group("arguments for evaluation")

        group.add_argument(
            '-cx',
            '--context',
            type=args_context,
            metavar='[cg]int',
            nargs='+',
            default=mx.cpu(),
            help='device context(s)')
        group.add_argument(
            '-db',
            '--dev_batch',
            type=int,
            metavar='int',
            default=2048,
            help='batch size for the development data')

        # arguments
        args = parser.parse_args(args)
        set_logger(args.log)

        args.vsm_list = [
            SimpleNamespace(
                model=n.type(
                    n.filepath),
                key=n.key) for n in args.vsm_list]
        if isinstance(args.context, list) and len(args.context) == 1:
            args.context = args.context[0]

        # component
        comp = TokenTagger(args.context, args.vsm_list)
        comp.load(args.model_path)

        # data
        dev_docs = args.reader.type(
            args.dev_path, args.reader.params, comp.key)

        # evaluate
        comp.evaluate(dev_docs, args.dev_batch)




# ========================================================================
# Copyright 2018 Emory University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
import abc
from typing import Any

from elit.util.structure import Document

__author__ = 'Jinho D. Choi'


class EvalMetric(abc.ABC):
    @abc.abstractmethod
    def update(self, document: Document):
        """
        Updates this evaluation metric given
        """
        pass

    @abc.abstractmethod
    def get(self) -> Any:
        """
        :return: the evaluated score.
        """
        pass


class Accuracy(EvalMetric):
    def __init__(self):
        self.correct = 0
        self.total = 0

    def __str__(self):
        return 'ACC:%6.2f (%d/%d)' % (self.get(), self.correct, self.total)

    @abc.abstractmethod
    def update(self, document: Document):
        pass

    def get(self) -> float:
        """
        :return: accuracy in percentage.
        """
        return 100.0 * self.correct / self.total if self.total > 0 else 0


class F1(EvalMetric):
    def __init__(self):
        super().__init__()
        self.correct = 0
        self.p_total = 0
        self.r_total = 0

    def __str__(self):
        return 'F1:%6.2f, P:%6.2f, R:%6.2f' % (
            self.f1(), self.precision(), self.recall())

    @abc.abstractmethod
    def update(self, document: Document):
        pass

    def precision(self) -> float:
        return 100.0 * self.correct / self.p_total if self.p_total > 0 else 0

    def recall(self) -> float:
        return 100.0 * self.correct / self.r_total if self.r_total > 0 else 0

    def f1(self):
        p = self.precision()
        r = self.recall()
        return 2 * p * r / (p + r) if p + r > 0 else 0

    def get(self) -> float:
        """
        :return: (F1 score, prediction, recall)
        """
        return self.f1()



# ========================================================================
# Copyright 2018 Emory University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
from typing import Any

import pytest

from elit.component import Component

__author__ = "Gary Lai"


def test_component():
    class TestComponent(Component):
        pass

    with pytest.raises(TypeError):
        TestComponent()


def test_space_tokenizer(space_tokenizer):
    result = space_tokenizer.decode("Hello, world")
    assert result['tok'] == ['Hello,', 'world']
    assert result['off'] == [(0, 6), (7, 12)]



# ========================================================================
# Copyright 2018 Emory University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
from typing import List

from elit.component import NLPComponent
from elit.doc_classifier import SentenceClassificationBatchState
from elit.eval import Accuracy, F1
from elit.util.structure import SENTI_NEUTRAL, SENTI_POSITIVE, SENTI_NEGATIVE

__author__ = 'Jinho D. Choi'


class SentimentState(SentenceClassificationBatchState):
    def eval(self, metric: List[Accuracy, Accuracy, F1, F1, F1]):
        eval_sentiment(metric, self.gold, [s[self.key] for s in self.document])


class SentimentAnalyzer(NLPComponent):
    def __init__(self, ctx, vsm):
        """
        :param ctx:
        :type ctx: mx.
        :param vsm:
        """
        super().__init__(ctx)
        self.vsm = vsm

        # to be initialized
        self.label_map = None


def eval_sentiment(metric: List[Accuracy,
                                Accuracy,
                                F1,
                                F1,
                                F1],
                   gold_labels: List[str],
                   pred_labels: List[str]):
    for i, gold in enumerate(gold_labels):
        pred = pred_labels[i]

        # accuracy: multi-class
        m: Accuracy = metric[0]
        m.total += 1
        if gold == pred:
            m.correct += 1

        # accuracy: binary
        if gold != SENTI_NEUTRAL:
            m: Accuracy = metric[1]
            m.total += 1
            if gold[0] == pred[0]:
                m.correct += 1

        # f1: positive, negative, neutral
        m: F1 = metric[2] if gold[0] == SENTI_POSITIVE else metric[3] if gold[0] == SENTI_NEGATIVE else metric[4]
        m.r_total += 1

        m: F1 = metric[2] if pred[0] == SENTI_POSITIVE else metric[3] if pred[0] == SENTI_NEGATIVE else metric[4]
        m.p_total += 1

        if gold[0] == pred[0]:
            m.correct += 1



# ========================================================================
# Copyright 2018 Emory University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
import abc
import argparse
import inspect
import logging
import re
import sys
from types import SimpleNamespace
from typing import Callable, Any, Dict, Tuple, Optional, Union, Type

import mxnet as mx

from elit.model import NLPModel
from elit.util.io import json_reader, tsv_reader
from elit.util.vsm import FastText, Word2Vec, VectorSpaceModel

__author__ = "Gary Lai, Jinho D. Choi"


class ComponentCLI(abc.ABC):
    """
    :class:`ComponentCLI` is an abstract class to implement a command-line interface for a component.

    Abstract methods to be implemented:
      - :meth:`ComponentCLI.train`
      - :meth:`ComponentCLI.decode`
      - :meth:`ComponentCLI.evaluate`
    """

    def __init__(self, name: str, description: str = None):
        """
        :param name: the name of the component.
        :param description: the description of this component; if ``None``, the name is used instead.
        """
        usage = [
            '        elit %s <command> [<args>]' %
            name,
            '',
            '    commands:',
            '           train: train a model (gold labels required)',
            '          decode: predict labels (gold labels not required)',
            '        evaluate: evaluate the pre-trained model (gold labels required)']

        usage = '\n'.join(usage)
        if not description:
            description = name
        parser = argparse.ArgumentParser(usage=usage, description=description)
        command = sys.argv[2]

        if not hasattr(self, command):
            logging.info('Unrecognized command: ' + command)
            parser.print_help()
            exit(1)
        getattr(self, command)(sys.argv[3:])

    @classmethod
    @abc.abstractmethod
    def train(cls, args):
        """
        :param args: the command-line arguments to be parsed by :class:`argparse.ArgumentParser`.

        Trains a model for this component.
        """
        raise NotImplementedError(
            '%s.%s()' %
            (cls.__class__.__name__, inspect.stack()[0][3]))

    @classmethod
    @abc.abstractmethod
    def decode(cls, args):
        """
        :param args: the command-line arguments to be parsed by :class:`argparse.ArgumentParser`.

        Predicts labels using this component.
        """
        raise NotImplementedError(
            '%s.%s()' %
            (cls.__class__.__name__, inspect.stack()[0][3]))

    @classmethod
    @abc.abstractmethod
    def evaluate(cls, args):
        """
        :param args: the command-line arguments to be parsed by :class:`argparse.ArgumentParser`.

        Evaluates the current model of this component.
        """
        raise NotImplementedError(
            '%s.%s()' %
            (cls.__class__.__name__, inspect.stack()[0][3]))


class ELITCLI(object):
    def __init__(self):
        parser = argparse.ArgumentParser(
            usage='''
    elit <command> [<args>]

commands:
    token_tagger    token_tagger
'''
        )
        parser.add_argument('command', help='command to run')
        args = parser.parse_args(sys.argv[1:2])
        if args.command == 'token_tagger':
            from elit.token_tagger import TokenTaggerCLI
            TokenTaggerCLI()
        else:
            print('Unrecognized command')
            parser.print_help()
            exit(1)


def set_logger(filename: str = None,
               level: int = logging.INFO,
               formatter: logging.Formatter = None):
    log = logging.getLogger()
    log.setLevel(level)
    ch = logging.StreamHandler(
        sys.stdout) if filename is None else logging.FileHandler(filename)
    if formatter is not None:
        ch.setFormatter(formatter)
    log.addHandler(ch)


def args_dict_str(s: str, const: Callable[[str], Any]) -> Dict[str, Any]:
    """
    :param s: key:int(,key:int)*
    :param const: type constructor (e.g., int, float)
    :return: a dictionary including key-value pairs from the input string.
    """
    r = re.compile(r'([A-Za-z0-9_-]+):([-]?[\d.]+)')
    d = {}

    for t in s.split(','):
        m = r.match(t)
        if m:
            d[m.group(1)] = const(m.group(2))
        else:
            raise argparse.ArgumentTypeError

    if not d:
        raise argparse.ArgumentTypeError
    return d


def args_dict_str_float(s: str) -> Dict[str, float]:
    return args_dict_str(s, float)


def args_tuple_int(s: str) -> Tuple[int, ...]:
    """
    :param s: int(, int)*
    :return: tuple of integers
    """
    r = re.compile(r'-?\d+(,-?\d+)*$')
    if r.match(s):
        return tuple(map(int, s.split(',')))
    else:
        raise argparse.ArgumentTypeError


def args_reader(s: str) -> SimpleNamespace:
    """
    :param s: json or tsv=(str:int)(,#1)*
    :return: the output of namespace_reader().
    """
    if s == 'json':
        return namespace_reader(reader_type=json_reader)
    if s.startswith('tsv='):
        return namespace_reader(reader_type=tsv_reader,
                                params=args_dict_str(s[4:], int))
    raise argparse.ArgumentTypeError


def args_vsm(s: str) -> SimpleNamespace:
    """
    :param s: (fasttext|word2vec):_key:filepath
    :return: the output of namespace_vsm().
    """
    i = s.find(':')
    if i < 1:
        raise argparse.ArgumentTypeError

    v = s[:i]
    if v == 'fasttext':
        vsm = FastText
    elif v == 'word2vec':
        vsm = Word2Vec
    else:
        raise argparse.ArgumentTypeError(
            "Unsupported vector space model: " + v)

    s = s[i + 1:]
    i = s.find(':')
    if i < 1:
        raise argparse.ArgumentTypeError

    key = s[:i]
    filepath = s[i + 1:]
    if not filepath:
        raise argparse.ArgumentTypeError

    return namespace_vsm(vsm_type=vsm, key=key, filepath=filepath)


def args_ngram_conv(s: str) -> Optional[SimpleNamespace]:
    """
    :param s: ngram:filters:activation:pool:dropout
    :return: the output of namespace_conf2d().
    """
    if s.lower() == 'none':
        return None
    c = s.split(':')
    pool = c[3] if c[3].lower() != 'none' else None
    return NLPModel.namespace_ngram_conv_layer(ngrams=args_tuple_int(c[0]),
                                               filters=int(c[1]),
                                               activation=c[2],
                                               pool=pool,
                                               dropout=float(c[4]))


def args_fuse_conv(s: str) -> Optional[SimpleNamespace]:
    """
    :param s: filters:activation:dropout
    :return: the output of namespace_conf2d().
    """
    if s.lower() == 'none':
        return None
    c = s.split(':')
    return NLPModel.namespace_fuse_conv_layer(filters=int(c[0]),
                                              activation=c[1],
                                              dropout=float(c[2]))


def args_hidden(s: str) -> Optional[SimpleNamespace]:
    """
    :param s: dim:activation:dropout
    :return: the output of namespace_hidden()
    """
    if s.lower() == 'none':
        return None
    c = s.split(':')
    return SimpleNamespace(dim=int(c[0]), activation=c[1], dropout=float(c[2]))


def args_context(s: str) -> mx.Context:
    """
    :param s: [cg]int
    :return: a device context
    """
    m = re.match(r'([cg])(\d*)', s)
    if m:
        d = int(m.group(2))
        return mx.cpu(d) if m.group(1) == 'c' else mx.gpu(d)
    else:
        raise argparse.ArgumentTypeError


def args_loss(s: str) -> mx.gluon.loss.Loss:
    s = s.lower()

    if s == 'softmaxcrossentropyloss':
        return mx.gluon.loss.SoftmaxCrossEntropyLoss()
    if s == 'sigmoidbinarycrossentropyloss':
        return mx.gluon.loss.SigmoidBinaryCrossEntropyLoss()
    if s == 'l2loss':
        return mx.gluon.loss.L2Loss()
    if s == 'l2loss':
        return mx.gluon.loss.L1Loss()
    if s == 'kldivloss':
        return mx.gluon.loss.KLDivLoss()
    if s == 'huberloss':
        return mx.gluon.loss.HuberLoss()
    if s == 'hingeloss':
        return mx.gluon.loss.HingeLoss()
    if s == 'squaredhingeloss':
        return mx.gluon.loss.SquaredHingeLoss()
    if s == 'logisticloss':
        return mx.gluon.loss.LogisticLoss()
    if s == 'tripletloss':
        return mx.gluon.loss.TripletLoss()
    if s == 'ctcloss':
        return mx.gluon.loss.CTCLoss()

    raise argparse.ArgumentTypeError("Unsupported loss: " + s)


def namespace_reader(reader_type: Union[json_reader,
                                        tsv_reader],
                     params: Optional[Dict] = None) -> SimpleNamespace:
    return SimpleNamespace(type=reader_type, params=params)


def namespace_vsm(
        vsm_type: Type[VectorSpaceModel],
        key: str,
        filepath: str) -> SimpleNamespace:
    return SimpleNamespace(type=vsm_type, key=key, filepath=filepath)


if __name__ == '__main__':
    ELITCLI()




# THIS FILE IS GENERATED FROM SCIPY SETUP.PY
short_version = '0.1.25'
version = '0.1.25'
full_version = '0.1.25.dev1535232511'
dev_version = '1535232511'
release = False
if not release:
    version = full_version



# ========================================================================
# Copyright 2018 Emory University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
import pytest

from elit.eval import Accuracy, F1
from elit.segmenter import EnglishSegmenter
from elit.tokenizer import Tokenizer, SpaceTokenizer, EnglishTokenizer

__author__ = "Gary Lai"


@pytest.fixture()
def tokenizer():
    return Tokenizer()


@pytest.fixture()
def space_tokenizer():
    return SpaceTokenizer()


@pytest.fixture()
def english_tokenizer():
    return EnglishTokenizer()


@pytest.fixture()
def english_segmenter():
    return EnglishSegmenter()


@pytest.fixture()
def accuracy():
    return Accuracy()


@pytest.fixture()
def f1():
    return F1()



# ========================================================================
# Copyright 2018 Emory University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
import abc
import codecs
import os
import re

from pkg_resources import resource_filename

from elit.component import Component
from elit.util.string import *
from elit.util.structure import TOK, OFF

__author__ = "Jinho D. Choi, Gary Lai"


class Tokenizer(Component):
    def __init__(self):
        super(Tokenizer, self).__init__()
        pass

    def init(self):
        pass

    def load(self, model_path, **kwargs):
        """

        :param model_path:
        :return:
        """
        pass

    def save(self, model_path, **kwargs):
        """

        :param model_path:
        :param kwargs:
        :return:
        """
        pass

    def train(self, trn_data, dev_data, model_path, **kwargs):
        """

        :param trn_data:
        :param dev_data:
        :param model_path:
        :return:
        """
        pass

    @abc.abstractmethod
    def decode(self, data, offset=0, **kwargs):
        """

        :type data: str
        :type offset: int
        :param data: the input text.
        :param offset: the starting offset.
        :return: the tuple of (tokens, offsets[, custom values]*); see the comments for Tokenizer.offsets() for more details about the offsets.
        :rtype: json
        """
        pass

    def evaluate(self, data, **kwargs):
        """

        :param data:
        :return:
        """
        pass

    @staticmethod
    def offsets(input_data, tokens, offset=0):
        """

        :type input_data: str
        :type tokens: list of str
        :type offset: int
        :param input_data: the input text.
        :param tokens: the list of tokens split from the input text.
        :param offset: offset of tokens
        :return: the list of (begin, end) offsets, where the begin (inclusive) and the end
        (exclusive) offsets indicate the caret positions of the first and the last characters of the
        corresponding token, respectively. e.g., text = 'Hello, world!',
        tokens = ['Hello', ',', 'world', '!'] -> [(0, 5), (5, 6), (7, 12), (12, 13)]
        :rtype: json
        """

        def get_offset(token):
            nonlocal end
            begin = input_data.index(token, end)
            end = begin + len(token)
            return begin + offset, end + offset

        end = 0
        return [get_offset(token) for token in tokens]


class SpaceTokenizer(Tokenizer):

    def __init__(self):
        super(SpaceTokenizer, self).__init__()

    def decode(self, input_data, offset=0, **kwargs):
        """

        :param input_data:
        :param offset:
        :param kwargs:
        :return:
        """
        tokens = input_data.split()

        return {
            TOK: tokens,
            OFF: self.offsets(input_data, tokens, offset)
        }


class EnglishTokenizer(Tokenizer):

    def __init__(self):
        super(EnglishTokenizer, self).__init__()
        self.ABBREVIATION_PERIOD = self.read_word_set(resource_filename(
            'elit.resources.tokenizer', 'english_abbreviation_period.txt'))
        self.APOSTROPHE_FRONT = self.read_word_set(resource_filename(
            'elit.resources.tokenizer', 'english_apostrophe_front.txt'))
        self.MAP_CONCAT_WORD = self.read_concat_word_dict(
            resource_filename(
                'elit.resources.tokenizer',
                'english_concat_words.txt'))
        self.HYPHEN_PREFIX = self.read_word_set(
            resource_filename(
                'elit.resources.tokenizer',
                'english_hyphen_prefix.txt'))
        self.HYPHEN_SUFFIX = self.read_word_set(
            resource_filename(
                'elit.resources.tokenizer',
                'english_hyphen_suffix.txt'))
        # regular expressions
        self.RE_NETWORK_PROTOCOL = re.compile(
            r"((http|https|ftp|sftp|ssh|ssl|telnet|smtp|pop3|imap|imap4|sip)(://))")
        """
        # :abc:
        # <3 </3 <\3
        # (: ): \: *: $: (-: (^: (= (;
        # :) :( =) B) 8) :-) :^) :3 :D :p :| :(( :---)
        """
        self.RE_EMOTICON = re.compile(
            r"(:\w+:|<[\\/]?3|[\(\)\\\|\*\$][-\^]?[:=\;]|[:\=\;B8]([-\^]+)?[3DOPp\@\$\*\(\)\\/\|]+)(\W|$)")
        """
        jinho@elit.cloud
        jinho.choi@elit.cloud
        choi@demo.elit.cloud
        jinho:choi@127.0.0.1
        """
        self.RE_EMAIL = re.compile(
            r"[\w\-\.]+(:\S+)?@(([A-Za-z0-9\-]+\.)+[A-Za-z]{2,12}|\d{1,3}(\.\d{1,3}){3})")
        """
        &arrow;
        &#123; &#x123; &#X123;
        """
        self.RE_HTML_ENTITY = re.compile(r"&([A-Za-z]+|#[Xx]?\d+);")
        """
        # [1] (1a) {A} <a1> [***] [A.a] [A.1] [1.a] ((---))
        """
        self.RE_LIST_ITEM = re.compile(
            r"(([\[\(\{\<]+)(\d+[A-Za-z]?|[A-Za-z]\d*|\W+)(\.(\d+|[A-Za-z]))*([\]\)\}\>])+)")
        """
        don't don’t I'll HE'S
        """
        self.RE_APOSTROPHE = re.compile(
            r"(?i)[a-z](n['\u2019]t|['\u2019](ll|nt|re|ve|[dmstz]))(\W|$)")
        """
        a.b.c 1-2-3
        """
        self.RE_ABBREVIATION = re.compile(r"[A-Za-z0-9]([\.-][A-Za-z0-9])*$")
        """
        10kg 1cm
        """
        self.RE_UNIT = re.compile(
            r"(?i)(\d)([acdfkmnpyz]?[mg]|[ap]\.m|ch|cwt|d|drc|ft|fur|gr|h|in|lb|lea|mi|ms|oz|pg|qtr|yd)$")
        """
        hello.World
        """
        self.RE_FINAL_MARK_IN_BETWEEN = re.compile(
            r"([A-Za-z]{3,})([\.\?\!]+)([A-Za-z]{3,})$")

    def load(self, model_path, *args, **kwargs):
        self.ABBREVIATION_PERIOD = self.read_word_set(
            os.path.join(model_path, 'english_abbreviation_period.txt'))
        self.APOSTROPHE_FRONT = self.read_word_set(
            os.path.join(model_path, 'english_apostrophe_front.txt'))
        self.MAP_CONCAT_WORD = self.read_concat_word_dict(
            os.path.join(model_path, 'english_concat_words.txt'))
        self.HYPHEN_PREFIX = self.read_word_set(
            os.path.join(model_path, 'english_hyphen_prefix.txt'))
        self.HYPHEN_SUFFIX = self.read_word_set(
            os.path.join(model_path, 'english_hyphen_suffix.txt'))

    def decode(self, input_data, offset=0, **kwargs):
        tokens = []
        offsets = []
        result = {
            TOK: tokens,
            OFF: offsets
        }
        # no valid token in the input text
        if not input_data or input_data.isspace():
            return result

        # skip beginning and ending spaces
        begin = next(i for i, c in enumerate(input_data) if not c.isspace())
        last = len(input_data) - next(
            i for i, c in enumerate(reversed(input_data)) if not c.isspace())

        # search for in-between spaces
        for end, c in enumerate(input_data[begin + 1:last], begin + 1):
            if c.isspace():
                self.tokenize_aux(tokens, offsets, input_data,
                                  begin, end, offset)
                begin = end + 1

        self.tokenize_aux(tokens, offsets, input_data, begin, last, offset)
        return result

    def tokenize_aux(self, tokens, offsets, text, begin, end, offset):
        if begin >= end or end > len(text):
            return False
        token = text[begin:end]

        # handle special cases
        if self.tokenize_trivial(tokens, offsets, token, begin, end, offset):
            return True
        if self.tokenize_regex(
                tokens,
                offsets,
                text,
                begin,
                end,
                offset,
                token):
            return True
        if self.tokenize_symbol(
                tokens,
                offsets,
                text,
                begin,
                end,
                offset,
                token):
            return True

        # add the token as it is
        self.add_token(tokens, offsets, token, begin, end, offset)
        return True

    def tokenize_trivial(self, tokens, offsets, token, begin, end, offset):
        if end - begin == 1 or token.isalnum():
            self.add_token(tokens, offsets, token, begin, end, offset)
            return True

        return False

    def tokenize_regex(self, tokens, offsets, text, begin, end, offset, token):
        def group(regex, gid=0):
            m = regex.search(token)
            if m:
                idx = begin + m.start(gid)
                lst = begin + m.end(gid)

                self.tokenize_aux(tokens, offsets, text, begin, idx, offset)
                self.add_token(tokens, offsets, m.group(gid), idx, lst, offset)
                self.tokenize_aux(tokens, offsets, text, lst, end, offset)
                return True

            return False

        def hyperlink():
            m = self.RE_NETWORK_PROTOCOL.search(token)
            if m:
                if m.start() > 0:
                    idx = begin + m.start()
                    self.tokenize_aux(tokens, offsets, text,
                                      begin, idx, offset)
                    self.add_token(tokens, offsets,
                                   token[m.start():], idx, end, offset)
                else:
                    self.add_token(tokens, offsets, token, begin, end, offset)
                return True

            return False

        # split by regular expressions
        if group(self.RE_HTML_ENTITY):
            return True
        if group(self.RE_EMAIL):
            return True
        if hyperlink():
            return True
        if group(self.RE_EMOTICON, 1):
            return True
        if group(self.RE_LIST_ITEM):
            return True
        if group(self.RE_APOSTROPHE, 1):
            return True
        return False

    def tokenize_symbol(
            self,
            tokens,
            offsets,
            text,
            begin,
            end,
            offset,
            token):
        def index_last_sequence(i, c):
            final_mark = is_final_mark(c)

            for j, d in enumerate(token[i + 1:], i + 1):
                if final_mark:
                    if not is_final_mark(d):
                        return j
                elif c != d:
                    return j

            return len(token)

        def skip(i, c):
            if c == '.' or c == '+':  # .1, +1
                return self.is_digit(token, i + 1)
            if c == '-':  # -1
                return i == 0 and self.is_digit(token, i + 1)
            if c == ',':  # 1,000,000
                return self.is_digit(
                    token,
                    i -
                    1) and self.is_digit(
                    token,
                    i +
                    1,
                    i +
                    4) and not self.is_digit(
                    token,
                    i +
                    4)
            if c == ':':
                # 1:2
                return self.is_digit(
                    token,
                    i -
                    1) and self.is_digit(
                    token,
                    i +
                    1)
            if is_single_quote(c):
                return self.is_digit(
                    token,
                    i +
                    1,
                    i +
                    3) and not self.is_digit(
                    token,
                    i +
                    3)  # '97
            return False

        def split(i, c, p0, p1):
            if p0(c):
                j = index_last_sequence(i, c)

                if p1(i, j):
                    idx = begin + i
                    lst = begin + j

                    self.tokenize_aux(tokens, offsets, text,
                                      begin, idx, offset)
                    self.add_token(tokens, offsets,
                                   token[i:j], idx, lst, offset)
                    self.tokenize_aux(tokens, offsets, text, lst, end, offset)
                    return True

            return False

        def separator_0(c):
            return c in {',', ';', ':', '~', '&', '|', '/'} or \
                is_bracket(c) or is_arrow(
                    c) or is_double_quote(c) or is_hyphen(c)

        def edge_symbol_0(c):
            return is_single_quote(c) or is_final_mark(c)

        def currency_like_0(c):
            return c == '#' or is_currency(c)

        def edge_symbol_1(i, j):
            return i + 1 < j or i == 0 or j == len(token) or is_punct(token[i - 1]) or is_punct(
                token[j])

        def currency_like_1(i, j):
            return i + 1 < j or j == len(token) or token[j].isdigit()

        # split by symbols
        for i, c in enumerate(token):
            if skip(i, c):
                continue
            if split(i, c, separator_0, lambda i, j: True):
                return True
            if split(i, c, edge_symbol_0, edge_symbol_1):
                return True
            if split(i, c, currency_like_0, currency_like_1):
                return True
        return False

    def add_token(self, tokens, offsets, token, begin, end, offset):
        if not self.concat_token(
                tokens,
                offsets,
                token,
                end) and not self.split_token(
                tokens,
                offsets,
                token,
                begin,
                end,
                offset):
            self.add_token_aux(tokens, offsets, token, begin, end, offset)

    def concat_token(self, tokens, offsets, token, end):
        def apostrophe_front(prev, curr):
            return len(prev) == 1 and is_single_quote(
                prev) and curr in self.APOSTROPHE_FRONT

        def abbreviation(prev, curr):
            return curr == '.' and (self.RE_ABBREVIATION.match(
                prev) or prev in self.ABBREVIATION_PERIOD)

        def acronym(prev, curr, next):
            return len(curr) == 1 and curr in {'&', '|', '/'} and (
                len(prev) <= 2 and len(next) <= 2 or prev.isupper() and next.isupper())

        def hyphenated(prev, curr, next):
            p = len(prev)

            if len(curr) == 1 and is_hyphen(curr):
                if self.is_digit(prev, p - 3, p) and (p == 3 or is_hyphen(
                        prev[p - 4])) and next.isdigit():
                    # 000-0000, 000-000-0000
                    return True
                if prev[-1].isalnum() and (len(prev) == 1 or is_hyphen(prev[p - 2])
                                           ) and len(next) == 1 and next.isalnum():
                    # p-u-s-h
                    return True
                return (prev in self.HYPHEN_PREFIX and next.isalnum()) or (
                    next in self.HYPHEN_SUFFIX and prev.isalnum())

            return False

        def no_dot_digit(prev, curr, next):
            if prev == 'no' and curr == '.' and next[0].isdigit():
                t, o = tokens.pop(), offsets.pop()
                tokens[-1] += t
                offsets[-1] = (offsets[-1][0], o[1])
                return True

            return False

        # concatenate split tokens if necessary
        if tokens:
            prev = tokens[-1].lower()
            curr = token.lower()

            if apostrophe_front(prev, curr) or abbreviation(prev, curr):
                tokens[-1] += token
                offsets[-1] = (offsets[-1][0], end)
                return True

        if len(tokens) >= 2:
            prev = tokens[-2].lower()
            curr = tokens[-1].lower()
            next = token.lower()

            if acronym(tokens[-2],
                       curr,
                       token) or hyphenated(prev,
                                            curr,
                                            next):
                tokens[-2] += tokens[-1] + token
                offsets[-2] = (offsets[-2][0], end)
                del tokens[-1]
                del offsets[-1]
                return True

            no_dot_digit(prev, curr, next)

        return False

    def split_token(self, tokens, offsets, token, begin, end, offset):
        def unit():
            m = self.RE_UNIT.search(token)
            if m:
                idx = begin + m.start(2)
                self.add_token_aux(
                    tokens, offsets, token[:m.start(2)], begin, idx, offset)
                self.add_token_aux(
                    tokens, offsets, m.group(2), idx, end, offset)
                return True
            return False

        def concat_words():
            t = self.MAP_CONCAT_WORD.get(token.lower(), None)
            if t:
                i = 0
                for j in t:
                    self.add_token_aux(
                        tokens, offsets, token[i:j], begin + i, begin + j, offset)
                    i = j
                return True
            return False

        def final_mark():
            m = self.RE_FINAL_MARK_IN_BETWEEN.match(token)
            if m:
                for i in range(1, 4):
                    self.add_token_aux(
                        tokens,
                        offsets,
                        m.group(i),
                        begin + m.start(i),
                        begin + m.end(i),
                        offset)
                return True
            return False

        return unit() or concat_words() or final_mark()

    @staticmethod
    def read_word_set(filename):
        fin = codecs.open(filename, mode='r', encoding='utf-8')
        s = set(line.strip() for line in fin)
        print('Init: %s(keys=%d)' % (filename, len(s)))
        return s

    @staticmethod
    def read_concat_word_dict(filename):
        def key_value(line):
            l = [i for i, c in enumerate(line) if c == ' ']
            l = [i - o for o, i in enumerate(l)]
            line = line.replace(' ', '')
            l.append(len(line))
            return line, l

        fin = codecs.open(filename, mode='r', encoding='utf-8')
        d = dict(key_value(line.strip()) for line in fin)
        print('Init: %s(keys=%d)' % (filename, len(d)))
        return d

    @staticmethod
    def add_token_aux(tokens, offsets, token, begin, end, offset):
        tokens.append(token)
        offsets.append((begin + offset, end + offset))

    @staticmethod
    def is_digit(token, i, j=None):
        if 0 <= i < len(token):
            if j is None:
                return token[i].isdigit()
            if i < j <= len(token):
                return token[i:j].isdigit()
        return False



# ========================================================================
# Copyright 2017 Emory University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
import abc
import inspect
import logging
import time
from typing import Union, Sequence, Dict, Any

import mxnet as mx
from mxnet import nd, gluon, autograd
from mxnet.ndarray import NDArray

from elit.eval import EvalMetric
from elit.util.iterator import BatchIterator, NLPIterator
from elit.util.structure import Document

__author__ = 'Jinho D. Choi, Gary Lai'


class Component(abc.ABC):
    """
    :class:`Component` is an abstract class; any component developed in ELIT must inherit this class.

    Abstract methods to be implemented:
      - :meth:`Component.init`
      - :meth:`Component.load`
      - :meth:`Component.save`
      - :meth:`Component.train`
      - :meth:`Component.decode`
      - :meth:`Component.evaluate`
    """

    @abc.abstractmethod
    def init(self, **kwargs):
        """
        :param kwargs: custom parameters.

        Initializes this component.
        """
        raise NotImplementedError(
            '%s.%s()' %
            (self.__class__.__name__, inspect.stack()[0][3]))

    @abc.abstractmethod
    def load(self, model_path: str, **kwargs):
        """
        :param model_path: the filepath where a model can be loaded.

        Loads a model for this component from the filepath.
        """
        raise NotImplementedError(
            '%s.%s()' %
            (self.__class__.__name__, inspect.stack()[0][3]))

    @abc.abstractmethod
    def save(self, model_path: str, **kwargs):
        """
        :param model_path: the filepath where the current model can be saved.

        Saves the current model of this component to the filepath.
        """
        raise NotImplementedError(
            '%s.%s()' %
            (self.__class__.__name__, inspect.stack()[0][3]))

    @abc.abstractmethod
    def train(
            self,
            trn_data: Any,
            dev_data: Any,
            model_path: str,
            **kwargs) -> float:
        """
        :param trn_data: training data.
        :param dev_data: development (validation) data.
        :param model_path: the filepath where the trained model(s) are to be saved.
        :return: the best score form the development data.

        Trains a model for this component and saves the model to the filepath.
        """
        raise NotImplementedError(
            '%s.%s()' %
            (self.__class__.__name__, inspect.stack()[0][3]))

    @abc.abstractmethod
    def decode(self, data: Any, **kwargs):
        """
        :param data: input data.

        Processes the input data, make predictions, and saves the predicted labels back to the input data.
        """
        raise NotImplementedError(
            '%s.%s()' %
            (self.__class__.__name__, inspect.stack()[0][3]))

    @abc.abstractmethod
    def evaluate(self, data: Any, **kwargs):
        """
        :param data: input data.

        Evaluates the current model of this component with the input data.
        """
        raise NotImplementedError(
            '%s.%s()' %
            (self.__class__.__name__, inspect.stack()[0][3]))


class NLPComponent(Component):
    """
    :class:`NLPComponent` is an abstract class; any NLP component developed in ELIT must inherit this class.
    It is similar to :class:`Component` except that the type of training and development data is specified to :class:`elit.structure.Document`.

    Abstract methods to be implemented:
      - :meth:`Component.init`
      - :meth:`Component.load`
      - :meth:`Component.save`
      - :meth:`NLPComponent.train`
      - :meth:`NLPComponent.decode`
      - :meth:`NLPComponent.evaluate`
    """

    @abc.abstractmethod
    def train(
            self,
            trn_docs: Sequence[Document],
            dev_docs: Sequence[Document],
            model_path: str,
            **kwargs) -> float:
        """
        :param trn_docs: the sequence of documents for training.
        :param dev_docs: the sequence of documents for development (validation).
        :param model_path: the filepath where trained model(s) are to be saved.
        :return: the best score form the development data.

        Trains a model for this component and saves the model to the filepath.
        """
        raise NotImplementedError(
            '%s.%s()' %
            (self.__class__.__name__, inspect.stack()[0][3]))

    @abc.abstractmethod
    def decode(self, docs: Sequence[Document], **kwargs):
        """
        :param docs: the sequence of input documents.

        Processes the input documents, make predictions, and saves the predicted labels back to the input documents.
        """
        raise NotImplementedError(
            '%s.%s()' %
            (self.__class__.__name__, inspect.stack()[0][3]))

    @abc.abstractmethod
    def evaluate(self, docs: Sequence[Document], **kwargs):
        """
        :param docs: the sequence of input documents.

        Evaluates the current model of this component with the input documents.
        """
        raise NotImplementedError(
            '%s.%s()' %
            (self.__class__.__name__, inspect.stack()[0][3]))


class MXNetComponent(NLPComponent):
    """
    :class:`MXNetComponent` is an abstract class to implement NLP components using MXNet for machine learning.
    :meth:`MXNetComponent.decode` and :meth:`MXNetComponent.train` are defined in this class.

    Abstract methods to be implemented:
      - :meth:`Component.init`
      - :meth:`Component.load`
      - :meth:`Component.save`
      - :meth:`MXNetComponent.data_iterator`
      - :meth:`MXNetComponent.eval_metric`
    """

    def __init__(self, ctx: Union[mx.Context, Sequence[mx.Context]]):
        """
        :param ctx: the (list of) device context(s) for :class:`mxnet.gluon.Block`.
        """
        self.ctx = ctx
        self.model = None

    @abc.abstractmethod
    def data_iterator(
            self,
            documents: Sequence[Document],
            batch_size: int,
            shuffle: bool,
            label: bool,
            **kwargs) -> NLPIterator:
        """
        :param documents: the sequence of input documents.
        :param batch_size: the size of batches to process at a time.
        :param shuffle: if ``True``, shuffle instances for every epoch; otherwise, no shuffle.
        :param label: if ``True``, each instance is a tuple of (feature vector, label); otherwise, it is just a feature vector.
        :return: the iterator to retrieve batches of training or decoding instances.
        """
        raise NotImplementedError(
            '%s.%s()' %
            (self.__class__.__name__, inspect.stack()[0][3]))

    @abc.abstractmethod
    def eval_metric(self, **kwargs) -> EvalMetric:
        """
        :return: the evaluation metric for this component.
        """
        raise NotImplementedError(
            '%s.%s()' %
            (self.__class__.__name__, inspect.stack()[0][3]))

    # override
    def train(self,
              trn_docs: Sequence[Document],
              dev_docs: Sequence[Document],
              model_path: str,
              trn_batch_size: int = 64,
              dev_batch_size: int = 2048,
              epoch: int = 100,
              loss: gluon.loss.Loss = None,
              optimizer: str = 'adagrad',
              optimizer_params: Dict[str, float] = None,
              **kwargs) -> float:
        """
        :param trn_docs: the sequence of documents for training.
        :param dev_docs: the sequence of documents for development (validation).
        :param model_path: the filepath where the trained model to be saved.
        :param trn_batch_size: the batch size applied to the training data.
        :param dev_batch_size: the batch size applied to the development data.
        :param epoch: the maximum number of epochs.
        :param loss: the `loss function <https://mxnet.incubator.apache.org/api/python/gluon/loss.html?highlight=softmaxcrossentropyloss#gluon-loss-api>`_; if not specified, ``SoftmaxCrossEntropyLoss`` is used.
        :param optimizer: the type of the `optimizer <https://mxnet.incubator.apache.org/api/python/optimization/optimization.html?highlight=optimiz#the-mxnet-optimizer-package>`_ for training.
        :param optimizer_params: the parameters for the optimizer.

        Trains and saves a model for this component.
        """
        log = ('Configuration',
               '- context(s): %s' % str(self.ctx),
               '- batch size: %d' % trn_batch_size,
               '- max epoch : %d' % epoch,
               '- loss func : %s' % str(loss),
               '- optimizer : %s <- %s' % (optimizer, optimizer_params))
        logging.info('\n'.join(log))

        # create iterators
        logging.info('Iterators')

        st = time.time()
        trn_iterator = self.data_iterator(
            trn_docs, trn_batch_size, shuffle=True, label=True, **kwargs)
        et = time.time()
        logging.info('- trn: %s (%d sec)' % (str(trn_iterator), et - st))

        st = time.time()
        dev_iterator = self.data_iterator(
            dev_docs,
            dev_batch_size,
            shuffle=False,
            label=True,
            **kwargs)
        et = time.time()
        logging.info('- dev: %s (%d sec)' % (str(dev_iterator), et - st))

        # create a trainer
        if loss is None:
            loss = gluon.loss.SoftmaxCrossEntropyLoss()
        trainer = gluon.Trainer(
            self.model.collect_params(),
            optimizer,
            optimizer_params)

        # train
        logging.info('Training')
        best_e, best_eval = -1, -1

        for e in range(1, epoch + 1):
            st = time.time()
            trn_acc = self.train_iter(trn_iterator, loss, trainer)
            mt = time.time()
            dev_metric = self.evaluate_iter(dev_iterator)
            et = time.time()

            ev = dev_metric.get()
            if best_eval < ev:
                best_e, best_eval = e, ev
                self.save(model_path)

            logging.info(
                '%4d: trn-time: %d, dev-time: %d, trn-acc: %6.2f, dev-eval: %5.2f, best-dev: %5.2f @%4d' %
                (e, mt - st, et - mt, trn_acc, dev_metric.get(), best_eval, best_e))

        return best_eval

    def train_iter(self,
                   iterator: NLPIterator,
                   loss: gluon.loss.Loss,
                   trainer: gluon.Trainer) -> float:
        """
        :param iterator: the iterator to retrieve batches of (feature vectors, labels).
        :param loss: the loss function.
        :param trainer: the trainer.
        :return: the evaluation metric including evaluation statistics on the training states.

        Trains all batches in the iterator for one epoch.
        """
        device = self.train_multiple_devices if isinstance(
            self.ctx, list) else self.train_single_device
        return device(iterator, loss, trainer)

    def train_single_device(self,
                            iterator: NLPIterator,
                            loss: gluon.loss.Loss,
                            trainer: gluon.Trainer) -> float:
        """
        :param iterator: the iterator to retrieve batches of (feature vectors, labels).
        :param loss: the loss function.
        :param trainer: the trainer.
        :return: the training accuracy.

        Trains all batches in the iterator with a single device.
        """
        state_begin, total, correct = 0, 0, 0
        others = []

        for batch in iterator:
            xs, ys = zip(*batch)
            total += len(ys)
            xs = nd.array(
                xs, self.ctx) if iterator.x1 else [
                nd.array(
                    x, self.ctx) for x in zip(
                    *xs)]
            ys = nd.array(ys, self.ctx)

            with autograd.record():
                t = self.model(xs) if iterator.x1 else self.model(*xs)
                if isinstance(t, NDArray):
                    output = t
                else:
                    output = t[0]
                    others = t[1:]

                l = loss(output, ys)
                l.backward()

            trainer.step(xs.shape[0])
            iterator.process(output, *others)
            correct += len([1 for o,
                            y in zip(mx.ndarray.argmax(output,
                                                       axis=1),
                                     ys) if int(o.asscalar()) == int(y.asscalar())])

        return 100.0 * correct / total

    def train_multiple_devices(self,
                               iterator: NLPIterator,
                               loss: gluon.loss.Loss,
                               trainer: gluon.Trainer) -> float:
        """
        :param iterator: the iterator to retrieve batches of (feature vectors, labels).
        :param loss: the loss function.
        :param trainer: the trainer.
        :return: the training accuracy.

        Trains all batches in the iterator with a single device.
        """

        def train():
            with autograd.record():
                ts = [
                    self.model(x_split) if iterator.x1 else self.model(
                        *x_split) for x_split in x_splits]
                if isinstance(ts[0], NDArray):
                    outputs = ts
                    others = empty
                else:
                    outputs = [t[0] for t in ts]
                    others = [t[1:] for t in ts]

                losses = [loss(output, y_split)
                          for output, y_split in zip(outputs, y_splits)]
                for l in losses:
                    l.backward()

            trainer.step(
                sum(x_split.shape[0] for x_split in x_splits), ignore_stale_grad=True)

            c, i = 0, 0
            for output, y_split in zip(outputs, y_splits):
                if others:
                    iterator.process(output, *others[i])
                else:
                    iterator.process(output)
                c += len([1 for o, y in zip(mx.ndarray.argmax(output, axis=1),
                                            y_split) if int(o.asscalar()) == int(y.asscalar())])
                i += 1
            return c

        x_splits, y_splits, empty = [], [], []
        total, correct = 0, 0

        for batch in iterator:
            if batch:
                xs, ys = zip(*batch)
                total += len(ys)
                ctx = self.ctx[len(x_splits)]
                x_splits.append(
                    nd.array(
                        xs, ctx) if iterator.x1 else [
                        nd.array(
                            x, ctx) for x in zip(
                            *xs)])
                y_splits.append(nd.array(ys, ctx))

            if len(x_splits) == len(self.ctx) or not batch:
                correct += train()
                x_splits, y_splits = [], []

        if x_splits:
            correct += train()
        return 100.0 * correct / total

    # override
    def decode(
            self,
            docs: Sequence[Document],
            batch_size: int = 2048,
            **kwargs):
        """
        :param docs: the sequence of input documents.
        :param batch_size: the batch size.

        Processes the input documents and saves the predicted labels to the input documents.
        """
        log = ('Decoding',
               '- context(s): %s' % str(self.ctx),
               '- batch size: %d' % batch_size)
        logging.info('\n'.join(log))

        iterator = self.data_iterator(
            docs, batch_size, shuffle=False, label=False)
        st = time.time()
        self.decode_iter(iterator)
        et = time.time()
        logging.info('- time: %d (sec)' % (et - st))

    def decode_iter(self, iterator: NLPIterator):
        """
        :param iterator: the iterator to retrieve batches of feature vectors.

        Decodes all batches in the iterator.
        """
        device = self.decode_multiple_devices if isinstance(
            self.ctx, list) else self.decode_single_device
        device(iterator)

    def decode_single_device(self, iterator: NLPIterator):
        """
        :param iterator: the iterator to retrieve batches of feature vectors.

        Decodes all batches in the iterator with a single device.
        """
        others = []

        for batch in iterator:
            if not isinstance(batch[0], NDArray):
                batch, _ = zip(*batch)
            xs = nd.array(batch, self.ctx)
            t = self.model(xs)
            if isinstance(t, NDArray):
                output = t
            else:
                output = t[0]
                others = t[1:]

            iterator.process(output, *others)

    def decode_multiple_devices(self, iterator: BatchIterator):
        """
        :param iterator: the iterator to retrieve batches of feature vectors.

        Decodes all batches in the iterator with multiple devices.
        """

        def decode():
            ts = [
                self.model(split) if iterator.x1 else self.model(
                    *split) for split in splits]
            if isinstance(ts[0], NDArray):
                outputs = ts
                others = empty
            else:
                outputs = [t[0] for t in ts]
                others = [t[1:] for t in ts]

            for i, output in enumerate(outputs):
                if others:
                    iterator.process(output, *others[i])
                else:
                    iterator.process(output)

        splits, empty = [], []

        for batch in iterator:
            if batch:
                if iterator.label:
                    batch, _ = zip(*batch)
                ctx = self.ctx[len(splits)]
                splits.append(
                    nd.array(
                        batch,
                        ctx) if iterator.x1 else [
                        nd.array(
                            x,
                            ctx) for x in zip(
                            *batch)])

            if len(splits) == len(self.ctx) or not batch:
                decode()
                splits = []

        if splits:
            decode()

    # override
    def evaluate(
            self,
            docs: Sequence[Document],
            batch_size: int = 2048,
            **kwargs):
        """
        :param docs: the sequence of input documents.
        :param batch_size: the batch size.
        :param kwargs: custom parameters.

        Evaluates the current model with the input documents.
        """
        log = ('Evaluating',
               '- context(s): %s' % str(self.ctx),
               '- batch size: %d' % batch_size)
        logging.info('\n'.join(log))

        iterator = self.data_iterator(
            docs, batch_size, shuffle=False, label=True)
        st = time.time()
        metric = self.evaluate_iter(iterator)
        et = time.time()
        logging.info('- time: %d (sec)' % (et - st))
        logging.info('%s' % str(metric))

    def evaluate_iter(self, iterator: NLPIterator, decode=True) -> EvalMetric:
        """
        :param iterator: the iterator to retrieve batches of (feature vectors, labels).
        :param decode: if ``True``, decodes all the states in the iterator before evaluation.
        :return: the evaluation metric including evaluation statistics on the input states.

        Evaluates all batches in the iterator.
        """
        if decode:
            self.decode_iter(iterator)
        metric = self.eval_metric()
        for state in iterator.states:
            metric.update(state.document)
        return metric



# ========================================================================
# Copyright 2018 Emory University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
# -*- coding: utf-8 -*-
"""
.. module:: elitsdk.benchmark
   :synopsis: The ELIT software development kit.
.. moduleauthor:: Gary Lai
"""
import timeit

__author__ = "Gary Lai"


class Timer(object):
    """
    Timer is an context manager measures elapsed time of running function or process.

    Example::

        from elitsdk.benchmark import Timer
        from example.example import SpaceTokenizer
        space_tokenizer = SpaceTokenizer()
        with Timer() as t:
            space_tokenizer.decode("hello world")
        print(t.runtime)

    """

    def __init__(self, timer=timeit.default_timer):
        self.timer = timer
        self.start = None
        self.end = None
        self.runtime = None

    def __enter__(self):
        self.start = self.timer()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.end = self.timer()
        self.runtime = self.end - self.start



#!/usr/bin/env python

from __future__ import division, print_function

import os
import subprocess
import sys

from setuptools import setup, find_packages

__author__ = "Gary Lai"

MAJOR = 0
MINOR = 1
MICRO = 25
ISRELEASED = False
VERSION = '%d.%d.%d' % (MAJOR, MINOR, MICRO)


# Return the git revision as a string
def get_dev_version():
    def _minimal_ext_cmd(cmd):
        # construct minimal environment
        env = {}
        for k in ['SYSTEMROOT', 'PATH', 'HOME']:
            v = os.environ.get(k)
            if v is not None:
                env[k] = v
        # LANGUAGE is used on win32
        env['LANGUAGE'] = 'C'
        env['LANG'] = 'C'
        env['LC_ALL'] = 'C'
        out = subprocess.Popen(cmd, stdout=subprocess.PIPE, env=env).communicate()[0]
        return out

    try:
        # out = _minimal_ext_cmd(['git', 'rev-parse', 'HEAD'])
        out = _minimal_ext_cmd(['git', 'log', '-1', '--pretty=format:%ct'])
        DEV_VERSION = out.strip().decode()
    except OSError:
        DEV_VERSION = "Unknown"

    return DEV_VERSION


def get_version_info(filename):
    # Adding the git rev number needs to be done inside write_version_py(),
    # otherwise the import of numpy.version messes up the build under Python 3.
    FULLVERSION = VERSION
    if os.path.exists('.git'):
        DEV_VERSION = get_dev_version()
    elif os.path.exists(filename):
        # must be a source distribution, use existing version file
        try:
            from elit.version import dev_version as DEV_VERSION
        except ImportError:
            raise ImportError("Unable to import dev_version. Try removing " \
                              "elit/version.py and the build directory " \
                              "before building.")
    else:
        DEV_VERSION = "Unknown"

    if not ISRELEASED:
        FULLVERSION += '.dev' + DEV_VERSION

    return FULLVERSION, DEV_VERSION


def write_version_py(filename):
    cnt = """
# THIS FILE IS GENERATED FROM SCIPY SETUP.PY
short_version = '%(version)s'
version = '%(version)s'
full_version = '%(full_version)s'
dev_version = '%(dev_version)s'
release = %(isrelease)s
if not release:
    version = full_version
"""
    FULLVERSION, DEV_VERSION  = get_version_info(filename)

    a = open(filename, 'w')
    try:
        a.write(cnt % {'version': VERSION,
                       'full_version': FULLVERSION,
                       'dev_version': DEV_VERSION ,
                       'isrelease': str(ISRELEASED)})
    finally:
        a.close()

VERSION_PY = 'elit/version.py'

if sys.version_info[:2] < (3, 5):
    raise RuntimeError("Python version >= 3.5 required.")

# How mature is this project? Common values are
#   3 - Alpha
#   4 - Beta
#   5 - Production/Stable
CLASSIFIERS = """\
Development Status :: 3 - Alpha
Intended Audience :: Science/Research
Intended Audience :: Developers
License :: OSI Approved :: Apache Software License
Programming Language :: Python :: 3 :: Only
Programming Language :: Python :: 3.5
Programming Language :: Python :: 3.6
Topic :: Scientific/Engineering :: Artificial Intelligence
Topic :: Scientific/Engineering :: Information Analysis
Operating System :: POSIX
Operating System :: Unix
Operating System :: MacOS
"""

EXCLUDE_FROM_PACKAGES = ['']


def setup_package():
    write_version_py(filename=VERSION_PY)

    metadata = dict(
        name='elit',
        url='https://github.com/elitcloud/elit',
        download_url='https://github.com/elitcloud/elit/tree/master',
        author='Jinho D. Choi',
        author_email='jinho.choi@emory.edu',
        description='The Emory Language Information Toolkit (ELIT).',
        license='ALv2',
        packages=find_packages(exclude=EXCLUDE_FROM_PACKAGES),
        package_data={'': ['*.txt']},
        install_requires=[
            'argparse==1.4.0',
            'pybind11==2.2.3',
            'yafasttext==0.8.22.1',
            'gensim==3.5.0',
            'numpy==1.14.5',
            'ujson==1.35',
            'mxnet==1.2.0',
        ],
        tests_require=[
            'pytest',
        ],
        classifiers=[_f for _f in CLASSIFIERS.split('\n') if _f],
        scripts=['bin/elit'],
    )
    metadata['version'] = get_version_info(filename=VERSION_PY)[0]

    setup(**metadata)


if __name__ == '__main__':
    setup_package()
